# -*- coding: utf-8 -*-

from collections import (
    defaultdict,
)
import numpy as np

from skyllh.core.dataset import (
    Dataset,
    DatasetData,
)
from skyllh.core.progressbar import (
    ProgressBar,
)
from skyllh.core.py import (
    classname,
    issequenceof,
)
from skyllh.core.source_hypo_grouping import (
    SourceHypoGroupManager,
)


class DetSigYieldService(
        object):
    """This class provides a service to build and hold detector signal yield
    instances for multiple datasets and source hypothesis groups.
    """

    def __init__(
            self,
            shg_mgr,
            dataset_list,
            data_list,
            ppbar=None,
            **kwargs):
        """Creates a new DetSigYieldService instance.
        """
        super().__init__(
            **kwargs)

        self._set_shg_mgr(shg_mgr)

        self.dataset_list = dataset_list
        self.data_list = data_list

        self._arr = self.construct_detsigyield_array(
            ppbar=ppbar)

    @property
    def shg_mgr(self):
        """(read-only) The instance of SourceHypoGroupManager providing the list
        of source hypothesis groups.
        """
        return self._shg_mgr

    @property
    def dataset_list(self):
        """The list of instance of Dataset for which the detector signal yields
        should be built.
        """
        return self._dataset_list

    @dataset_list.setter
    def dataset_list(self, datasets):
        if not issequenceof(datasets, Dataset):
            raise TypeError(
                'The dataset_list property must be a sequence of Dataset '
                'instances! '
                f'Its current type is {classname(datasets)}!')
        self._dataset_list = list(datasets)

    @property
    def data_list(self):
        """The list of instance of DatasetData for which the detector signal
        yields should be built.
        """
        return self._data_list

    @data_list.setter
    def data_list(self, datas):
        if not issequenceof(datas, DatasetData):
            raise TypeError(
                'The data_list property must be a sequence of DatasetData '
                'instances! '
                f'Its current type is {classname(datas)}!')
        self._data_list = list(datas)

    @property
    def arr(self):
        """(read-only) The (N_datasets, N_source_hypo_groups)-shaped numpy
        ndarray of object holding the constructed DetSigYield instances.
        """
        return self._arr

    @property
    def n_datasets(self):
        """(read-only) The number of datasets this service was created for.
        """
        return self._arr.shape[0]

    @property
    def n_shgs(self):
        """(read-only) The number of source hypothesis groups this service was
        created for.
        """
        return self._arr.shape[1]

    def _set_shg_mgr(self, mgr):
        """Sets the internal member variable to the given instance of
        SourceHypoGroupManager.
        """
        if not isinstance(mgr, SourceHypoGroupManager):
            raise TypeError(
                'The shg_mgr argument must be an instance of '
                'SourceHypoGroupManager! '
                f'Its current type is {classname(mgr)}!')

        self._shg_mgr = mgr

    def change_shg_mgr(
            self,
            shg_mgr,
            ppbar=None,
    ):
        """Changes the instance of SourceHypoGroupManager of this service. This
        will also rebuild the detector signal yields.
        """
        self._set_shg_mgr(shg_mgr)

        self._arr = self.construct_detsigyield_array(
            ppbar=ppbar)

    def get_builder_to_shgidxs_dict(
            self,
            ds_idx,
    ):
        """Creates a dictionary with the builder instance as key and the list of
        source hypo group indices to which the builder applies as value.
        Hence, SHGs using the same builder instance can be grouped for
        DetSigYield construction.

        Parameters
        ----------
        ds_idx : int
            The index of the dataset for which the same builders apply.

        Returns
        -------
        builder_shgidxs_dict : dict
            The dictionary with the builder instance as key and the list of
            source hypo group indices to which the builder applies as value.
        """
        n_datasets = len(self._dataset_list)

        if ds_idx < 0 or ds_idx >= n_datasets:
            raise ValueError(
                f'The dataset index {ds_idx} must be within the range '
                f'[0,{n_datasets-1}]!')

        builder_shgidxs_dict = defaultdict(list)
        for (g, shg) in enumerate(self._shg_mgr.shg_list):

            builder_list = shg.detsigyield_builder_list
            if (len(builder_list) != 1) and (len(builder_list) != n_datasets):
                raise ValueError(
                    'The number of detector signal yield builders '
                    f'({len(builder_list)}) is not 1 and does not '
                    f'match the number of datasets ({n_datasets}) for the '
                    f'{g}th source hypothesis group!')

            builder = (
                builder_list[0] if len(builder_list) == 1 else
                builder_list[ds_idx]
            )

            builder_shgidxs_dict[builder].append(g)

        return builder_shgidxs_dict

    def construct_detsigyield_array(
            self,
            ppbar=None,
    ):
        """Creates a (N_datasets, N_source_hypo_groups)-shaped numpy ndarray of
        object holding the constructed DetSigYield instances.

        If the same DetSigYieldBuilder class is used for all source hypotheses
        of a particular dataset, the
        :meth:`~skyllh.core.detsigyield.DetSigYieldBuilder.construct_detsigyields`
        method is called with different flux models to optimize the construction
        of the detector signal yield functions.

        Parameters
        ----------
        ppbar : instance of ProgressBar | None
            The instance of ProgressBar of the optional parent progress bar.

        Returns
        -------
        detsigyield_arr : instance of numpy.ndarray
            The (N_datasets, N_source_hypo_groups)-shaped numpy ndarray of
            object holding the constructed DetSigYield instances.
        """
        n_datasets = len(self._dataset_list)

        detsigyield_arr = np.empty(
            (n_datasets,
             self._shg_mgr.n_src_hypo_groups),
            dtype=object
        )

        pbar = ProgressBar(
            self._shg_mgr.n_src_hypo_groups * n_datasets,
            parent=ppbar).start()

        shg_list = self.shg_mgr.shg_list

        for (j, (dataset, data)) in enumerate(zip(self._dataset_list,
                                                  self._data_list)):

            builder_to_shgidxs_dict = self.get_builder_to_shgidxs_dict(ds_idx=j)

            for (builder, shgidxs) in builder_to_shgidxs_dict.items():
                factory = builder.get_detsigyield_construction_factory()
                if factory is None:
                    # The builder does not provide a factory for DetSigYield
                    # instance construction. So we have to construct the
                    # detector signal yields one by one for each SHG.
                    for g in shgidxs:
                        shg = shg_list[g]

                        detsigyield = builder.construct_detsigyield(
                            dataset=dataset,
                            data=data,
                            shg=shg,
                            ppbar=pbar)

                        detsigyield_arr[j, g] = detsigyield

                        pbar.increment()
                else:
                    # The builder provides a factory for the construction of
                    # several DetSigYield instances simultaneously, one for each
                    # flux model.
                    shgs = [
                        shg_list[g]
                        for g in shgidxs
                    ]

                    detsigyields = factory(
                        dataset=dataset,
                        data=data,
                        shgs=shgs,
                        ppbar=pbar)

                    for (i, g) in enumerate(shgidxs):
                        detsigyield_arr[j, g] = detsigyields[i]

                    pbar.increment(len(detsigyields))

        pbar.finish()

        return detsigyield_arr


class SrcDetSigYieldWeightsService(
        object):
    r"""This class provides a service for the source detector signal yield
    weights, which are the product of the source weights with the detector
    signal yield, denoted with :math:`a_{j,k}(\vec{p}_{\mathrm{s}_k})` in the
    math formalism documentation.

    .. math::

        a_{j,k}(\vec{p}_{\mathrm{s}_k}) = W_k
            \mathcal{Y}_{\mathrm{s}_{j,k}}(\vec{p}_{\mathrm{s}_k})

    The service has a method to calculate the weights and a method to retrieve
    the weights. The weights are stored internally.
    """

    @staticmethod
    def create_src_recarray_list_list(
            detsigyield_service,
    ):
        """Creates a list of numpy record ndarrays, one for each source
        hypothesis group suited for evaluating the detector signal yield
        instance of that source hypothesis group.

        Parameters
        ----------
        detsigyield_service : instance of DetSigYieldService
            The instance of DetSigYieldService providing the
            (N_datasets, N_source_hypo_groups)-shaped 2D ndarray of
            DetSigYield instances, one for each dataset and source hypothesis
            group combination.

        Returns
        -------
        src_recarray_list_list : list of list of numpy record ndarrays
            The (N_datasets,N_source_hypo_groups)-shaped list of list of the
            source numpy record ndarrays, one for each dataset and source
            hypothesis group combination, which is needed for
            evaluating a particular detector signal yield instance.
        """
        n_datasets = detsigyield_service.n_datasets
        n_shgs = detsigyield_service.n_shgs
        shg_list = detsigyield_service.shg_mgr.shg_list

        src_recarray_list_list = []
        for ds_idx in range(n_datasets):
            src_recarray_list = []
            for shg_idx in range(n_shgs):
                shg = shg_list[shg_idx]
                src_recarray_list.append(
                    detsigyield_service.arr[ds_idx][shg_idx].sources_to_recarray(
                        shg.source_list))

            src_recarray_list_list.append(src_recarray_list)

        return src_recarray_list_list

    @staticmethod
    def create_src_weight_array_list(
            shg_mgr,
    ):
        """Creates a list of numpy 1D ndarrays holding the source weights, one
        for each source hypothesis group.

        Parameters
        ----------
        shg_mgr : instance of SourceHypoGroupManager
            The instance of SourceHypoGroupManager defining the source
            hypothesis groups with their sources.

        Returns
        -------
        src_weight_array_list : list of numpy 1D ndarrays
            The list of 1D numpy ndarrays holding the source weights, one for
            each source hypothesis group.
        """
        src_weight_array_list = [
            np.array([src.weight for src in shg.source_list])
            for shg in shg_mgr.shg_list
        ]
        return src_weight_array_list

    def __init__(
            self,
            detsigyield_service,
            **kwargs,
    ):
        """Creates a new SrcDetSigYieldWeightsService instance.

        Parameters
        ----------
        detsigyield_service : instance of DetSigYieldService
            The instance of DetSigYieldService providing the
            (N_datasets, N_source_hypo_groups)-shaped array of DetSigYield
            instances, one instance for each combination of dataset and source
            hypothesis group.
        """
        super().__init__(
            **kwargs)

        self.detsigyield_service = detsigyield_service

        # Create the list of list of source record arrays for each combination
        # of dataset and source hypothesis group.
        self._src_recarray_list_list = type(self).create_src_recarray_list_list(
            detsigyield_service=self._detsigyield_service)

        # Create the list of 1D ndarrays holding the source weights for each
        # source hypothesis group.
        self._src_weight_array_list = type(self).create_src_weight_array_list(
            shg_mgr=self._detsigyield_service.shg_mgr)

        self._a_jk = None
        self._a_jk_grads = None

    @property
    def shg_mgr(self):
        """(read-only) The instance of SourceHypoGroupManager defining the
        source hypothesis groups.
        """
        return self._detsigyield_service.shg_mgr

    @property
    def detsigyield_service(self):
        """The instance of DetSigYieldService providing the
        (N_datasets, N_source_hypo_groups)-shaped array of DetSigYield
        instances.
        """
        return self._detsigyield_service

    @detsigyield_service.setter
    def detsigyield_service(self, service):
        if not isinstance(service, DetSigYieldService):
            raise TypeError(
                'The detsigyield_service property must be an instance of '
                'DetSigYieldService! '
                f'Its current type is {classname(service)}!')
        self._detsigyield_service = service

    @property
    def detsigyield_arr(self):
        """(read-only) The (N_datasets, N_source_hypo_groups)-shaped 2D numpy
        ndarray holding the DetSigYield instances for each source hypothesis
        group.
        """
        return self._detsigyield_service.arr

    @property
    def n_datasets(self):
        """(read-only) The number of datasets this service was created for.
        """
        return self._detsigyield_service.n_datasets

    @property
    def n_shgs(self):
        """(read-only) The number of source hypothesis groups this service was
        created for.
        """
        return self._detsigyield_service.n_shgs

    @property
    def src_recarray_list_list(self):
        """(read-only) The (N_datasets,N_source_hypo_groups)-shaped list of list
        of the source numpy record ndarrays, one for each dataset and source
        hypothesis group combination, which is needed for evaluating a
        particular detector signal yield instance.
        """
        return self._src_recarray_list_list

    def change_shg_mgr(
            self,
            shg_mgr,
    ):
        """Re-creates the internal source numpy record arrays needed for the
        detector signal yield calculation.

        Parameters
        ----------
        shg_mgr : instance of SourceHypoGroupManager
            The new SourceHypoGroupManager instance.
        """
        if id(shg_mgr) != id(self._detsigyield_service.shg_mgr):
            raise ValueError(
                'The provides instance of SourceHypoGroupManager does not '
                'match the instance of the detector signal yield service!')

        self._src_recarray_list_list = type(self).create_src_recarray_list_list(
            detsigyield_service=self._detsigyield_service)

        self._src_weight_array_list = type(self).create_src_weight_array_list(
            shg_mgr=self._detsigyield_service.shg_mgr)

    def calculate(
            self,
            src_params_recarray):
        """Calculates the source detector signal yield weights for each source
        and their derivative w.r.t. each global floating parameter. The result
        is stored internally as:

            a_jk : instance of ndarray
                The (N_datasets,N_sources)-shaped numpy ndarray holding the
                source detector signal yield weight for each combination of
                dataset and source.
            a_jk_grads : dict
                The dictionary holding the (N_datasets,N_sources)-shaped numpy
                ndarray with the derivatives w.r.t. the global fit parameter
                the SrcDetSigYieldWeightsService depend on. The dictionary's key
                is the index of the global fit parameter.

        Parameters
        ----------
        src_params_recarray : instance of numpy record ndarray
            The numpy record ndarray of length N_sources holding the local
            source parameters. See the documentation of
            :meth:`skyllh.core.parameters.ParameterModelMapper.create_src_params_recarray`
            for more information about this record array.
        """
        n_datasets = self.n_datasets

        shg_mgr = self._detsigyield_service.shg_mgr

        self._a_jk = np.empty(
            (n_datasets, shg_mgr.n_sources,),
            dtype=np.double)

        self._a_jk_grads = defaultdict(
            lambda: np.zeros(
                (n_datasets, shg_mgr.n_sources),
                dtype=np.double))

        sidx = 0
        for (shg_idx, (shg, src_weights)) in enumerate(zip(
                shg_mgr.shg_list, self._src_weight_array_list)):

            shg_n_src = shg.n_sources

            shg_src_slice = slice(sidx, sidx+shg_n_src)

            shg_src_params_recarray = src_params_recarray[shg_src_slice]

            for ds_idx in range(n_datasets):
                detsigyield = self._detsigyield_service.arr[ds_idx, shg_idx]
                src_recarray = self._src_recarray_list_list[ds_idx][shg_idx]

                (Yg, Yg_grads) = detsigyield(
                    src_recarray=src_recarray,
                    src_params_recarray=shg_src_params_recarray)

                self._a_jk[ds_idx][shg_src_slice] = src_weights * Yg

                for gpidx in Yg_grads.keys():
                    self._a_jk_grads[gpidx][ds_idx, shg_src_slice] =\
                        src_weights * Yg_grads[gpidx]

            sidx += shg_n_src

    def get_weights(self):
        """Returns the source detector signal yield weights and their
        derivatives w.r.t. the global fit parameters.

        Returns
        -------
        a_jk : instance of ndarray
            The (N_datasets, N_sources)-shaped numpy ndarray holding the
            source detector signal yield weight for each combination of
            dataset and source.
        a_jk_grads : dict
            The dictionary holding the (N_datasets, N_sources)-shaped numpy
            ndarray with the derivatives w.r.t. the global fit parameter
            the SrcDetSigYieldWeightsService depend on. The dictionary's key
            is the index of the global fit parameter.
        """
        return (self._a_jk, self._a_jk_grads)


class DatasetSignalWeightFactorsService(
        object):
    r"""This class provides a service to calculates the dataset signal weight
    factors, :math:`f_j(\vec{p}_\mathrm{s})`, for each dataset.
    It utilizes the source detector signal yield weights
    :math:`a_{j,k}(\vec{p}_{\mathrm{s}_k})`, provided by the
    :class:`~SrcDetSigYieldWeightsService` class.
    """

    def __init__(
            self,
            src_detsigyield_weights_service):
        r"""Creates a new DatasetSignalWeightFactors instance.

        Parameters
        ----------
        src_detsigyield_weights_service : instance of SrcDetSigYieldWeightsService
            The instance of SrcDetSigYieldWeightsService providing the source
            detector signal yield weights
            :math:`a_{j,k}(\vec{p}_{\mathrm{s}_k})`.
        """
        self.src_detsigyield_weights_service = src_detsigyield_weights_service

    @property
    def src_detsigyield_weights_service(self):
        r"""The instance of SrcDetSigYieldWeightsService providing the source
        detector signal yield weights :math:`a_{j,k}(\vec{p}_{\mathrm{s}_k})`.
        """
        return self._src_detsigyield_weights_service

    @src_detsigyield_weights_service.setter
    def src_detsigyield_weights_service(self, service):
        if not isinstance(service, SrcDetSigYieldWeightsService):
            raise TypeError(
                'The src_detsigyield_weights_service property must be an '
                'instance of SrcDetSigYieldWeightsService!')
        self._src_detsigyield_weights_service = service

    @property
    def n_datasets(self):
        """(read-only) The number of datasets.
        """
        return self._src_detsigyield_weights_service.n_datasets

    def calculate(self):
        r"""Calculates the dataset signal weight factors,
        :math:`f_j(\vec{p}_\mathrm{s})`. The result is stored internally as:

            f_j : instance of ndarray
                The (N_datasets,)-shaped 1D numpy ndarray holding the dataset
                signal weight factor for each dataset.
            f_j_grads : dict
                The dictionary holding the (N_datasets,)-shaped numpy
                ndarray with the derivatives w.r.t. the global fit parameter
                the DatasetSignalWeightFactorsService depend on.
                The dictionary's key is the index of the global fit parameter.
        """
        (a_jk, a_jk_grads) = self._src_detsigyield_weights_service.get_weights()

        a_j = np.sum(a_jk, axis=1)
        a = np.sum(a_jk)

        self._f_j = a_j / a

        # Calculate the derivative of f_j w.r.t. all floating parameters present
        # in the a_jk_grads using the quotient rule of differentiation.
        self._f_j_grads = dict()
        for gpidx in a_jk_grads.keys():
            # a is a scalar.
            # a_j is a (N_datasets)-shaped ndarray.
            # a_jk_grads is a dict of length N_gfl_params with values of
            #    (N_datasets,N_sources)-shaped ndarray.
            # a_j_grads is a (N_datasets,)-shaped ndarray.
            # a_grads is a scalar.
            a_j_grads = np.sum(a_jk_grads[gpidx], axis=1)
            a_grads = np.sum(a_jk_grads[gpidx])
            self._f_j_grads[gpidx] = (a_j_grads * a - a_j * a_grads) / a**2

    def get_weights(self):
        """Returns the

        Returns
        -------
        f_j : instance of ndarray
            The (N_datasets,)-shaped 1D numpy ndarray holding the dataset
            signal weight factor for each dataset.
        f_j_grads : dict
            The dictionary holding the (N_datasets,)-shaped numpy
            ndarray with the derivatives w.r.t. the global fit parameter
            the DatasetSignalWeightFactorsService depend on.
            The dictionary's key is the index of the global fit parameter.
        """
        return (self._f_j, self._f_j_grads)



# -*- coding: utf-8 -*-

from skyllh.core.background_generation import (
    BackgroundGenerationMethod,
)
from skyllh.core.py import (
    classname,
)
from skyllh.core.scrambling import (
    DataScrambler,
)


class FixedScrambledExpDataI3BkgGenMethod(
        BackgroundGenerationMethod,
):
    """This class implements the background event generation method for the
    IceCube detector using scrambled experimental data as background hypothesis
    with a fixed number of background events equal to the number of events in
    the dataset. This background generation method is the one used in SkyLab.
    """
    def __init__(
            self,
            data_scrambler,
            **kwargs,
    ):
        """Creates a new background generation method instance to generate
        background events from scrambled experimental data with a fixed number
        of events equal to the number of events in the dataset.

        Parameters
        ----------
        data_scrambler : instance of DataScrambler
            The DataScrambler instance to use to generate scrambled experimental
            data.
        """
        super().__init__(**kwargs)

        self.data_scrambler = data_scrambler

    @property
    def data_scrambler(self):
        """The DataScrambler instance that implements the data scrambling.
        """
        return self._data_scrambler

    @data_scrambler.setter
    def data_scrambler(self, scrambler):
        if not isinstance(scrambler, DataScrambler):
            raise TypeError(
                'The data_scrambler property must be an instance of '
                'DataScrambler! '
                f'Its current type is {classname(scrambler)}!')
        self._data_scrambler = scrambler

    def generate_events(
            self,
            rss,
            dataset,
            data,
            **kwargs,
    ):
        """Generates background events from the given data, by scrambling the
        experimental data. The number of events is equal to the size of the
        given dataset.

        Parameters
        ----------
        rss : instance of RandomStateService
            The instance of RandomStateService that should be used to generate
            random numbers from. It is used to scramble the experimental data.
        dataset : instance of Dataset
            The Dataset instance describing the dataset for which background
            events should get generated.
        data : instance of DatasetData
            The DatasetData instance holding the data of the dataset for which
            background events should get generated.

        Returns
        -------
        n_bkg : int
            The number of generated background events.
        bkg_events : instance of DataFieldRecordArray
            The instance of DataFieldRecordArray holding the generated
            background events.
        """
        bkg_events = self._data_scrambler.scramble_data(
            rss=rss,
            dataset=dataset,
            data=data.exp,
            copy=True)

        return (len(bkg_events), bkg_events)



# -*- coding: utf-8 -*-

from scipy.stats import (
    rv_continuous,
)

from skyllh.core.flux_model import (
    TimeFluxProfile,
)
from skyllh.core.py import (
    classname,
)


def create_scipy_stats_rv_continuous_from_TimeFluxProfile(
        profile,
):
    """This function builds a scipy.stats.rv_continuous instance for a given
    :class:`~skyllh.core.flux_model.TimeFluxProfile` instance.

    It can be used to generate random numbers according to the given time flux
    profile function.

    Parameters
    ----------
    profile : instance of TimeFluxProfile
        The instance of TimeFluxProfile providing the function of the time flux
        profile.

    Returns
    -------
    rv : instance of rv_continuous_frozen
        The instance of rv_continuous_frozen representing the time flux profile
        as a continuous random variate instance.
    """
    if not isinstance(profile, TimeFluxProfile):
        raise TypeError(
            'The profile argument must be an instance of TimeFluxProfile! '
            f'Its current type is {classname(profile)}!')

    norm = 0
    tot_integral = profile.get_total_integral()
    if tot_integral != 0:
        norm = 1 / tot_integral

    class rv_continuous_from_TimeFluxProfile(
            rv_continuous):

        def __init__(self, *args, **kwargs):
            """Creates a new instance of the subclass of rv_continuous using
            the time flux profile.
            """
            self._profile = profile
            self._norm = norm

            super().__init__(*args, **kwargs)

        def _pdf(self, t):
            """Calculates the probability density of the time flux profile
            function for given time values.
            """
            pd = self._profile(t=t) * self._norm

            return pd

        def _cdf(self, t):
            """Calculates the cumulative distribution function values for rhe
            given time values. If the time flux profile instance provides a
            ``cdf`` method, it will be used. Otherwise the generic ``_cdf``
            method of the ``rv_continuous`` class will be used.
            """
            if hasattr(self._profile, 'cdf') and callable(self._profile.cdf):
                return self._profile.cdf(t=t)

            return super()._cdf(t)

    rv = rv_continuous_from_TimeFluxProfile(
        a=profile.t_start,
        b=profile.t_stop,
    ).freeze(loc=0, scale=1)

    return rv



# -*- coding: utf-8 -*-

import abc
import copy
import pickle
import os.path
import sys

import numpy as np

from skyllh.core import (
    display as dsp,
    tool,
)
from skyllh.core.py import (
    classname,
    get_byte_size_prefix,
    getsizeof,
    issequence,
    issequenceof,
)


# Define a file loader registry that holds the FileLoader classes for different
# file formats.
_FILE_LOADER_REG = dict()


def register_FileLoader(
        formats,
        fileloader_cls):
    """Registers the given file formats (file extensions) to the given
    FileLoader class.

    Parameters
    ----------
    formats : str | list of str
        The list of file name extensions that should be mapped to the FileLoader
        class.
    fileloader_cls : instance of FileLoader
        The subclass of FileLoader that should be used for the given file
        formats.
    """
    if isinstance(formats, str):
        formats = [formats]
    if not issequence(formats):
        raise TypeError(
            'The "formats" argument must be a sequence!')
    if not issubclass(fileloader_cls, FileLoader):
        raise TypeError(
            'The "fileloader_cls" argument must be a subclass of FileLoader!')

    for fmt in formats:
        if fmt in _FILE_LOADER_REG.keys():
            raise KeyError(
                f'The format "{fmt}" is already registered!')
        _FILE_LOADER_REG[fmt] = fileloader_cls


def create_FileLoader(
        pathfilenames,
        **kwargs):
    """Creates the appropriate FileLoader object for the given file names.
    It looks up the FileLoader class from the FileLoader registry for the
    file name extension of the first file name in the given list.

    Parameters
    ----------
    pathfilenames : str | sequence of str
        The sequence of fully qualified file names of the files that should be
        loaded.

    Additional Parameters
    ---------------------
    Additional parameters will be passed to the constructor method of the
    chosen FileLoader class.

    Returns
    -------
    fileloader : FileLoader
        The appropriate FileLoader instance for the given type of data files.
    """
    if isinstance(pathfilenames, str):
        pathfilenames = [pathfilenames]
    if not issequenceof(pathfilenames, str):
        raise TypeError(
            'The pathfilenames argument must be a sequence of str!')

    # Sort the file names extensions with shorter extensions before longer ones
    # to support a format that is sub-string of another format.
    formats = sorted(_FILE_LOADER_REG.keys())
    for fmt in formats:
        fmt_len = len(fmt)
        if pathfilenames[0][-fmt_len:].lower() == fmt.lower():
            cls = _FILE_LOADER_REG[fmt]
            return cls(pathfilenames, **kwargs)

    raise RuntimeError(
        'No FileLoader class is suitable to load the data file '
        f'"{pathfilenames[0]}"!')


def assert_file_exists(
        pathfilename):
    """Checks if the given file exists and raises a RuntimeError if it does
    not exist.
    """
    if not os.path.isfile(pathfilename):
        raise RuntimeError(
            f'The data file "{pathfilename}" does not exist!')


class FileLoader(
        object,
        metaclass=abc.ABCMeta):
    """Abstract base class for a FileLoader class.
    """
    def __init__(
            self,
            pathfilenames,
            **kwargs):
        """Creates a new FileLoader instance.

        Parameters
        ----------
        pathfilenames : str | sequence of str
            The sequence of fully qualified file names of the data files that
            need to be loaded.
        """
        super().__init__(
            **kwargs)

        self.pathfilename_list = pathfilenames

    @property
    def pathfilename_list(self):
        """The list of fully qualified file names of the data files.
        """
        return self._pathfilename_list

    @pathfilename_list.setter
    def pathfilename_list(self, pathfilenames):
        if isinstance(pathfilenames, str):
            pathfilenames = [pathfilenames]
        if not issequence(pathfilenames):
            raise TypeError(
                'The pathfilename_list property must be of type str or a '
                'sequence of type str!')
        self._pathfilename_list = list(pathfilenames)

    @abc.abstractmethod
    def load_data(self, **kwargs):
        """This method is supposed to load the data from the file.
        """
        pass


class NPYFileLoader(
        FileLoader):
    """The NPYFileLoader class provides the data loading functionality for
    numpy data files containing numpy arrays. It uses the ``numpy.load``
    function for loading the data and the numpy.append function to concatenate
    several data files.
    """
    def __init__(
            self,
            pathfilenames,
            **kwargs):
        """Creates a new NPYFileLoader instance.

        Parameters
        ----------
        pathfilenames : str | sequence of str
            The sequence of fully qualified file names of the data files that
            need to be loaded.
        """
        super().__init__(
            pathfilenames=pathfilenames,
            **kwargs)

    def _load_file_memory_efficiently(
            self,
            pathfilename,
            keep_fields,
            dtype_conversions,
            dtype_conversion_except_fields):
        """Loads a single file in a memory efficient way.

        Parameters
        ----------
        pathfilename : str
            The fully qualified file name of the to-be-loaded file.
        keep_fields : list of str | None
            The list of field names which should be kept.

        Returns
        -------
        data : DataFieldRecordArray instance
            An instance of DataFieldRecordArray holding the data.
        """
        assert_file_exists(pathfilename)

        # Create a memory map into the data file. This loads the data only when
        # accessing the data.
        mmap_ndarray = np.load(pathfilename, mmap_mode='r')
        field_names = mmap_ndarray.dtype.names
        fname_to_fidx = dict([
            (fname, idx)
            for (idx, fname) in enumerate(field_names)
        ])
        dt_fields = mmap_ndarray.dtype.fields
        n_rows = mmap_ndarray.shape[0]

        data = dict()

        # Create empty arrays for each column of length n_rows.
        for fname in field_names:
            # Ignore fields that should not get kept.
            if (keep_fields is not None) and (fname not in keep_fields):
                continue

            # Get the original data type of the field.
            dt = dt_fields[fname][0]
            # Convert the data type if requested.
            if (fname not in dtype_conversion_except_fields) and\
               (dt in dtype_conversions):
                dt = dtype_conversions[dt]

            data[fname] = np.empty((n_rows,), dtype=dt)

        # Loop through the rows of the recarray.
        bs = 4096
        for ridx in range(0, n_rows):
            row = mmap_ndarray[ridx]
            for fname in data.keys():
                fidx = fname_to_fidx[fname]
                data[fname][ridx] = row[fidx]

            # Reopen the data file after each given blocksize.
            if ridx % bs == 0:
                del mmap_ndarray
                mmap_ndarray = np.load(pathfilename, mmap_mode='r')

        # Close the memory map file.
        del mmap_ndarray

        # Create a DataFieldRecordArray out of the dictionary.
        data = DataFieldRecordArray(data, copy=False)

        return data

    def _load_file_time_efficiently(
            self,
            pathfilename,
            keep_fields,
            dtype_conversions,
            dtype_conversion_except_fields):
        """Loads a single file in a time efficient way. This will load the data
        column-wise.
        """
        assert_file_exists(pathfilename)

        # Create a memory map into the data file. This loads the data only when
        # accessing the data.
        mmap_ndarray = np.load(pathfilename, mmap_mode='r')

        # Create a DataFieldRecordArray out of the memory mapped file. We need
        # to copy the data, otherwise we get read-only numpy arrays.
        data = DataFieldRecordArray(
            mmap_ndarray,
            keep_fields=keep_fields,
            dtype_conversions=dtype_conversions,
            dtype_conversion_except_fields=dtype_conversion_except_fields,
            copy=True)

        # Close the memory map file.
        del mmap_ndarray

        return data

    def load_data(  # noqa: C901
            self,
            keep_fields=None,
            dtype_conversions=None,
            dtype_conversion_except_fields=None,
            efficiency_mode=None):
        """Loads the data from the files specified through their fully qualified
        file names.

        Parameters
        ----------
        keep_fields : str | sequence of str | None
            Load the data into memory only for these data fields. If set to
            ``None``, all in-file-present data fields are loaded into memory.
        dtype_conversions : dict | None
            If not None, this dictionary defines how data fields of specific
            data types get converted into the specified data types.
            This can be used to use less memory.
        dtype_conversion_except_fields : str | sequence of str | None
            The sequence of field names whose data type should not get
            converted.
        efficiency_mode : str | None
            The efficiency mode the data should get loaded with. Possible values
            are:

                ``'memory'``:
                    The data will be load in a memory efficient way. This will
                    require more time, because all data records of a file will
                    be loaded sequentially.
                ``'time'``
                    The data will be loaded in a time efficient way. This will
                    require more memory, because each data file gets loaded in
                    memory at once.

            The default value is ``'time'``. If set to ``None``, the default
            value will be used.

        Returns
        -------
        data : instance of DataFieldRecordArray
            The DataFieldRecordArray holding the loaded data.

        Raises
        ------
        RuntimeError if a file does not exist.
        """
        if keep_fields is not None:
            if isinstance(keep_fields, str):
                keep_fields = [keep_fields]
            elif not issequenceof(keep_fields, str):
                raise TypeError(
                    'The keep_fields argument must be None, an instance of '
                    'type str, or a sequence of instances of type str!')

        if dtype_conversions is None:
            dtype_conversions = dict()
        elif not isinstance(dtype_conversions, dict):
            raise TypeError(
                'The dtype_conversions argument must be None, or an instance '
                'of dict!')

        if dtype_conversion_except_fields is None:
            dtype_conversion_except_fields = []
        elif isinstance(dtype_conversion_except_fields, str):
            dtype_conversion_except_fields = [dtype_conversion_except_fields]
        elif not issequenceof(dtype_conversion_except_fields, str):
            raise TypeError(
                'The dtype_conversion_except_fields argument must be a '
                'sequence of str instances.')

        efficiency_mode2func = {
            'memory': self._load_file_memory_efficiently,
            'time': self._load_file_time_efficiently
        }
        if efficiency_mode is None:
            efficiency_mode = 'time'
        if not isinstance(efficiency_mode, str):
            raise TypeError(
                'The efficiency_mode argument must be an instance of type str!')
        if efficiency_mode not in efficiency_mode2func:
            raise ValueError(
                'The efficiency_mode argument value must be one of '
                f'{", ".join(efficiency_mode2func.keys())}!')
        load_file_func = efficiency_mode2func[efficiency_mode]

        # Load the first data file.
        data = load_file_func(
            self._pathfilename_list[0],
            keep_fields=keep_fields,
            dtype_conversions=dtype_conversions,
            dtype_conversion_except_fields=dtype_conversion_except_fields
        )

        # Load possible subsequent data files by appending to the first data.
        for i in range(1, len(self._pathfilename_list)):
            data.append(load_file_func(
                self._pathfilename_list[i],
                keep_fields=keep_fields,
                dtype_conversions=dtype_conversions,
                dtype_conversion_except_fields=dtype_conversion_except_fields
            ))

        return data


class ParquetFileLoader(
        FileLoader
):
    """The ParquetFileLoader class provides the data loading functionality for
    parquet files. It uses the ``pyarrow`` package.
    """
    @tool.requires('pyarrow', 'pyarrow.parquet')
    def __init__(
            self,
            pathfilenames,
            **kwargs
    ):
        """Creates a new file loader instance for parquet data files.

        Parameters
        ----------
        pathfilenames : str | sequence of str
            The sequence of fully qualified file names of the data files that
            need to be loaded.
        """
        super().__init__(
            pathfilenames=pathfilenames,
            **kwargs)

        self.pa = tool.get('pyarrow')
        self.pq = tool.get('pyarrow.parquet')

    def load_data(
            self,
            keep_fields=None,
            dtype_conversions=None,
            dtype_conversion_except_fields=None,
            copy=False,
            **kwargs,
    ):
        """Loads the data from the files specified through their fully qualified
        file names.

        Parameters
        ----------
        keep_fields : str | sequence of str | None
            Load the data into memory only for these data fields. If set to
            ``None``, all in-file-present data fields are loaded into memory.
        dtype_conversions : dict | None
            If not ``None``, this dictionary defines how data fields of specific
            data types get converted into the specified data types.
            This can be used to use less memory.
        dtype_conversion_except_fields : str | sequence of str | None
            The sequence of field names whose data type should not get
            converted.
        copy : bool
            If set to ``True``, the column data from the pyarrow.Table instance
            will be copied into the DataFieldRecordArray. This should not be
            necessary.

        Returns
        -------
        data : instance of DataFieldRecordArray
            The DataFieldRecordArray holding the loaded data.
        """
        assert_file_exists(self.pathfilename_list[0])
        table = self.pq.read_table(self.pathfilename_list[0], columns=keep_fields)
        for pathfilename in self.pathfilename_list[1:]:
            assert_file_exists(pathfilename)
            next_table = self.pq.read_table(pathfilename, columns=keep_fields)
            table = self.pa.concat_tables([table, next_table])

        data = DataFieldRecordArray(
            data=table,
            data_table_accessor=ParquetDataTableAccessor(),
            keep_fields=keep_fields,
            dtype_conversions=dtype_conversions,
            dtype_conversion_except_fields=dtype_conversion_except_fields,
            copy=copy)

        return data


class PKLFileLoader(
        FileLoader):
    """The PKLFileLoader class provides the data loading functionality for
    pickled Python data files containing Python data structures. It uses the
    `pickle.load` function for loading the data from the file.
    """
    def __init__(
            self,
            pathfilenames,
            pkl_encoding=None,
            **kwargs):
        """Creates a new file loader instance for a pickled data file.

        Parameters
        ----------
        pathfilenames : str | sequence of str
            The sequence of fully qualified file names of the data files that
            need to be loaded.
        pkl_encoding : str | None
            The encoding of the pickled data files. If None, the default
            encodings 'ASCII' and 'latin1' will be tried to load the data.
        """
        super().__init__(
            pathfilenames=pathfilenames,
            **kwargs)

        self.pkl_encoding = pkl_encoding

    @property
    def pkl_encoding(self):
        """The encoding of the pickled data files. Can be None.
        If None, the default encodings 'ASCII' and 'latin1' will be tried to
        load the data.
        """
        return self._pkl_encoding

    @pkl_encoding.setter
    def pkl_encoding(self, encoding):
        if encoding is not None:
            if not isinstance(encoding, str):
                raise TypeError(
                    'The pkl_encoding property must be None or of type str!')
        self._pkl_encoding = encoding

    def load_data(
            self,
            **kwargs):
        """Loads the data from the files specified through their fully qualified
        file names.

        Returns
        -------
        data : Python object | list of Python objects
            The de-pickled Python object. If more than one file was specified,
            this is a list of Python objects, i.e. one object for each file.
            The file <-> object mapping order is preserved.

        Raises
        ------
        RuntimeError if a file does not exist.
        """
        # Define the possible encodings of the pickled files.
        encodings = ['ASCII', 'latin1']
        if self._pkl_encoding is not None:
            encodings = [self._pkl_encoding] + encodings

        data = []
        for pathfilename in self.pathfilename_list:
            assert_file_exists(pathfilename)
            with open(pathfilename, 'rb') as ifile:
                enc_idx = 0
                load_ok = False
                obj = None
                while (not load_ok) and (enc_idx < len(encodings)):
                    try:
                        encoding = encodings[enc_idx]
                        obj = pickle.load(ifile, encoding=encoding)
                    except UnicodeDecodeError:
                        enc_idx += 1
                        # Move the file pointer back to the beginning of the
                        # file.
                        ifile.seek(0)
                    else:
                        load_ok = True
                if obj is None:
                    raise RuntimeError(
                        f'The file "{pathfilename}" could not get unpickled! '
                        'No correct encoding available!')
                data.append(obj)

        if len(data) == 1:
            data = data[0]

        return data


class TextFileLoader(
        FileLoader):
    """The TextFileLoader class provides the data loading functionality for
    data text files where values are stored in a comma, or whitespace, separated
    format. It uses the numpy.loadtxt function to load the data. It reads the
    first line of the text file for a table header.
    """
    def __init__(
            self,
            pathfilenames,
            header_comment='#',
            header_separator=None,
            **kwargs):
        """Creates a new file loader instance for a text data file.

        Parameters
        ----------
        pathfilenames : str | sequence of str
            The sequence of fully qualified file names of the data files that
            need to be loaded.
        header_comment : str
            The character that defines a comment line in the text file.
        header_separator : str | None
            The separator of the header field names. If None, it assumes
            whitespaces.
        """
        super().__init__(
            pathfilenames=pathfilenames,
            **kwargs)

        self.header_comment = header_comment
        self.header_separator = header_separator

    @property
    def header_comment(self):
        """The character that defines a comment line in the text file.
        """
        return self._header_comment

    @header_comment.setter
    def header_comment(self, s):
        if not isinstance(s, str):
            raise TypeError(
                'The header_comment property must be of type str!')
        self._header_comment = s

    @property
    def header_separator(self):
        """The separator of the header field names. If None, it assumes
        whitespaces.
        """
        return self._header_separator

    @header_separator.setter
    def header_separator(self, s):
        if s is not None:
            if not isinstance(s, str):
                raise TypeError(
                    'The header_separator property must be None or of type '
                    'str!')
        self._header_separator = s

    def _extract_column_names(self, line):
        """Tries to extract the column names of the data table based on the
        given line.

        Parameters
        ----------
        line : str
            The text line containing the column names.

        Returns
        -------
        names : list of str | None
            The column names.
            It returns None, if the column names cannot be extracted.
        """
        # Remove possible new-line character and leading white-spaces.
        line = line.strip()
        # Check if the line is a comment line.
        if line[0:len(self._header_comment)] != self._header_comment:
            return None
        # Remove the leading comment character(s).
        line = line.strip(self._header_comment)
        # Remove possible leading whitespace characters.
        line = line.strip()
        # Split the line into the column names.
        names = line.split(self._header_separator)
        # Remove possible whitespaces of column names.
        names = [n.strip() for n in names]

        if len(names) == 0:
            return None

        return names

    def _load_file(
            self,
            pathfilename,
            keep_fields,
            dtype_conversions,
            dtype_conversion_except_fields):
        """Loads the given file.

        Parameters
        ----------
        pathfilename : str
            The fully qualified file name of the data file that
            need to be loaded.
        keep_fields : str | sequence of str | None
            Load the data into memory only for these data fields. If set to
            ``None``, all in-file-present data fields are loaded into memory.
        dtype_conversions : dict | None
            If not None, this dictionary defines how data fields of specific
            data types get converted into the specified data types.
            This can be used to use less memory.
        dtype_conversion_except_fields : str | sequence of str | None
            The sequence of field names whose data type should not get
            converted.

        Returns
        -------
        data : DataFieldRecordArray instance
            The DataFieldRecordArray instance holding the loaded data.
        """
        assert_file_exists(pathfilename)

        with open(pathfilename, 'r') as ifile:
            line = ifile.readline()
            column_names = self._extract_column_names(line)
            if column_names is None:
                raise ValueError(
                    f'The data text file "{pathfilename}" does not contain a '
                    'readable table header as first line!')
            usecols = None
            dtype = [(n, np.float64) for n in column_names]
            if keep_fields is not None:
                # Select only the given columns.
                usecols = []
                dtype = []
                for (idx, name) in enumerate(column_names):
                    if name in keep_fields:
                        usecols.append(idx)
                        dtype.append((name, np.float64))
                usecols = tuple(usecols)
            if len(dtype) == 0:
                raise ValueError(
                    'No data columns were selected to be loaded!')

            data_ndarray = np.loadtxt(
                ifile,
                dtype=dtype,
                comments=self._header_comment,
                usecols=usecols)

        data = DataFieldRecordArray(
            data_ndarray,
            keep_fields=keep_fields,
            dtype_conversions=dtype_conversions,
            dtype_conversion_except_fields=dtype_conversion_except_fields,
            copy=False)

        return data

    def load_data(
            self,
            keep_fields=None,
            dtype_conversions=None,
            dtype_conversion_except_fields=None,
            **kwargs):
        """Loads the data from the data files specified through their fully
        qualified file names.

        Parameters
        ----------
        keep_fields : str | sequence of str | None
            Load the data into memory only for these data fields. If set to
            ``None``, all in-file-present data fields are loaded into memory.
        dtype_conversions : dict | None
            If not None, this dictionary defines how data fields of specific
            data types get converted into the specified data types.
            This can be used to use less memory.
        dtype_conversion_except_fields : str | sequence of str | None
            The sequence of field names whose data type should not get
            converted.

        Returns
        -------
        data : instance of DataFieldRecordArray
            The DataFieldRecordArray holding the loaded data.

        Raises
        ------
        RuntimeError
            If a file does not exist.
        ValueError
            If the table header cannot be read.
        """
        if keep_fields is not None:
            if isinstance(keep_fields, str):
                keep_fields = [keep_fields]
            elif not issequenceof(keep_fields, str):
                raise TypeError(
                    'The keep_fields argument must be None, an instance of '
                    'type str, or a sequence of instances of type str!')

        if dtype_conversions is None:
            dtype_conversions = dict()
        elif not isinstance(dtype_conversions, dict):
            raise TypeError(
                'The dtype_conversions argument must be None, or an instance '
                'of dict!')

        if dtype_conversion_except_fields is None:
            dtype_conversion_except_fields = []
        elif isinstance(dtype_conversion_except_fields, str):
            dtype_conversion_except_fields = [dtype_conversion_except_fields]
        elif not issequenceof(dtype_conversion_except_fields, str):
            raise TypeError(
                'The dtype_conversion_except_fields argument must be a '
                'sequence of str instances.')

        # Load the first data file.
        data = self._load_file(
            self._pathfilename_list[0],
            keep_fields=keep_fields,
            dtype_conversions=dtype_conversions,
            dtype_conversion_except_fields=dtype_conversion_except_fields
        )

        # Load possible subsequent data files by appending to the first data.
        for i in range(1, len(self._pathfilename_list)):
            data.append(self._load_file(
                self._pathfilename_list[i],
                keep_fields=keep_fields,
                dtype_conversions=dtype_conversions,
                dtype_conversion_except_fields=dtype_conversion_except_fields
            ))

        return data


class DataTableAccessor(
        object,
        metaclass=abc.ABCMeta,
):
    """This class provides an interface wrapper to access the data table of a
    particular format in a unified way.
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @abc.abstractmethod
    def get_column(self, data, name):
        """This method is supposed to return a numpy.ndarray holding the data of
        the column with name ``name``.

        Parameters
        ----------
        data : any
            The data table.
        name : str
            The name of the column.

        Returns
        -------
        arr : instance of numpy.ndarray
            The column data as numpy ndarray.
        """
        pass

    @abc.abstractmethod
    def get_field_names(self, data):
        """This method is supposed to return a list of field names.
        """
        pass

    @abc.abstractmethod
    def get_field_name_to_dtype_dict(self, data):
        """This method is supposed to return a dictionary with field name and
        numpy dtype instance for each field.
        """
        pass

    @abc.abstractmethod
    def get_length(self, data):
        """This method is supposed to return the length of the data table.
        """
        pass


class NDArrayDataTableAccessor(
        DataTableAccessor,
):
    """This class provides an interface wrapper to access the data table stored
    as a structured numpy ndarray.
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get_column(self, data, name):
        """Gets the column data from the structured ndarray.

        Parameters
        ----------
        data : instance of numpy.ndarray
            The structured numpy ndarray holding the table data.
        name : str
            The name of the column.
        """
        return data[name]

    def get_field_names(self, data):
        return data.dtype.names

    def get_field_name_to_dtype_dict(self, data):
        """Returns the dictionary with field name and numpy dtype instance for
        each field.
        """
        fname_to_dtype_dict = dict([
            (k, v[0]) for (k, v) in data.dtype.fields.items()
        ])
        return fname_to_dtype_dict

    def get_length(self, data):
        """Returns the length of the data table.
        """
        length = data.shape[0]
        return length


class DictDataTableAccessor(
        DataTableAccessor,
):
    """This class provides an interface wrapper to access the data table stored
    as a Python dictionary.
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get_column(self, data, name):
        """Gets the column data from the dictionary.

        Parameters
        ----------
        data : dict
            The dictionary holding the table data.
        name : str
            The name of the column.
        """
        return data[name]

    def get_field_names(self, data):
        return list(data.keys())

    def get_field_name_to_dtype_dict(self, data):
        """Returns the dictionary with field name and numpy dtype instance for
        each field.
        """
        fname_to_dtype_dict = dict([
            (fname, data[fname].dtype) for fname in data.keys()
        ])
        return fname_to_dtype_dict

    def get_length(self, data):
        """Returns the length of the data table.
        """
        length = 0
        if len(data) > 0:
            length = data[next(iter(data))].shape[0]
        return length


class ParquetDataTableAccessor(
        DataTableAccessor,
):
    """This class provides an interface wrapper to access the data table stored
    as a Parquet table.
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get_column(self, data, name):
        """Gets the column data from the Parquet table.

        Parameters
        ----------
        data : instance of pyarrow.Table
            The instance of pyarrow.Table holding the table data.
        name : str
            The name of the column.
        """
        return data[name].to_numpy()

    def get_field_names(self, data):
        return data.column_names

    def get_field_name_to_dtype_dict(self, data):
        """Returns the dictionary with field name and numpy dtype instance for
        each field.
        """
        fname_to_dtype_dict = dict([
            (fname, data.field(fname).type.to_pandas_dtype())
            for fname in data.column_names
        ])
        return fname_to_dtype_dict

    def get_length(self, data):
        """Returns the length of the data table.
        """
        return len(data)


class DataFieldRecordArrayDataTableAccessor(
        DataTableAccessor,
):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get_column(self, data, name):
        """Gets the column data from the Parquet table.

        Parameters
        ----------
        data : instance of pyarrow.Table
            The instance of pyarrow.Table holding the table data.
        name : str
            The name of the column.
        """
        return data[name]

    def get_field_names(self, data):
        return data.field_name_list

    def get_field_name_to_dtype_dict(self, data):
        """Returns the dictionary with field name and numpy dtype instance for
        each field.
        """
        fname_to_dtype_dict = dict([
            (fname, data[fname].dtype)
            for fname in data.field_name_list
        ])
        return fname_to_dtype_dict

    def get_length(self, data):
        """Returns the length of the data table.
        """
        return len(data)


class DataFieldRecordArray(
        object):
    """The DataFieldRecordArray class provides a data container similar to a numpy
    record ndarray. But the data fields are stored as individual numpy ndarray
    objects. Hence, access of single data fields is much faster compared to
    access on the record ndarray.
    """
    def __init__(  # noqa: C901
            self,
            data,
            data_table_accessor=None,
            keep_fields=None,
            dtype_conversions=None,
            dtype_conversion_except_fields=None,
            copy=True,
    ):
        """Creates a DataFieldRecordArray from the given data.

        Parameters
        ----------
        data : any | None
            The tabulated data in any format. The only requirement is that
            there is a DataTableAccessor instance available for the given data
            format. Supported data types are:

                numpy.ndarray
                    A structured numpy ndarray.
                dict
                    A Python dictionary with field names as keys and
                    one-dimensional numpy.ndarrays as values.
                pyarrow.Table
                    An instance of pyarrow.Table.
                DataFieldRecordArray
                    An instance of DataFieldRecordArray.

            If set to `None`, the DataFieldRecordArray instance is initialized
            with no data and the length of the array is set to 0.
        data_table_accessor : instance of DataTableAccessor | None
            The instance of DataTableAccessor which provides column access to
            ``data``. If set to ``None``, an appropriate ``DataTableAccessor``
            instance will be selected based on the type of ``data``.
        keep_fields : str | sequence of str | None
            If not None (default), this specifies the data fields that should
            get kept from the given data. Otherwise all data fields get kept.
        dtype_conversions : dict | None
            If not None, this dictionary defines how data fields of specific
            data types get converted into the specified data types.
            This can be used to use less memory.
        dtype_conversion_except_fields : str | sequence of str | None
            The sequence of field names whose data type should not get
            converted.
        copy : bool
            Flag if the input data should get copied. Default is True. If a
            DataFieldRecordArray instance is provided, this option is set to
            ``True`` automatically.
        """
        self._data_fields = dict()
        self._len = None

        if data is None:
            data = dict()

        if keep_fields is not None:
            if isinstance(keep_fields, str):
                keep_fields = [keep_fields]
            elif not issequenceof(keep_fields, str):
                raise TypeError(
                    'The keep_fields argument must be None, an instance of '
                    'type str, or a sequence of instances of type str!')

        if dtype_conversions is None:
            dtype_conversions = dict()
        elif not isinstance(dtype_conversions, dict):
            raise TypeError(
                'The dtype_conversions argument must be None, or an instance '
                'of dict!')

        if dtype_conversion_except_fields is None:
            dtype_conversion_except_fields = []
        elif isinstance(dtype_conversion_except_fields, str):
            dtype_conversion_except_fields = [dtype_conversion_except_fields]
        elif not issequenceof(dtype_conversion_except_fields, str):
            raise TypeError(
                'The dtype_conversion_except_fields argument must be a '
                'sequence of str instances.')

        # Select an appropriate data table accessor for the type of data.
        if data_table_accessor is None:
            if isinstance(data, np.ndarray):
                data_table_accessor = NDArrayDataTableAccessor()
            elif isinstance(data, dict):
                data_table_accessor = DictDataTableAccessor()
            elif (tool.is_available('pyarrow') and
                  isinstance(data, tool.get('pyarrow').Table)):
                data_table_accessor = ParquetDataTableAccessor()
            elif isinstance(data, DataFieldRecordArray):
                data_table_accessor = DataFieldRecordArrayDataTableAccessor()
            else:
                raise TypeError(
                    'No TableDataAccessor instance has been specified for the '
                    f'data of type {type(data)}!')

        field_names = data_table_accessor.get_field_names(data)
        fname2dtype = data_table_accessor.get_field_name_to_dtype_dict(data)
        length = data_table_accessor.get_length(data)

        for fname in field_names:
            # Ignore fields that should not get kept.
            if (keep_fields is not None) and (fname not in keep_fields):
                continue

            copy_field = copy
            dt = fname2dtype[fname]
            if (fname not in dtype_conversion_except_fields) and\
               (dt in dtype_conversions):
                dt = dtype_conversions[dt]
                # If a data type conversion is needed, the data of the field
                # needs to get copied.
                copy_field = True

            if copy_field is True:
                # Create a ndarray with the final data type and then assign the
                # values from the data, which technically is a copy.
                field_arr = np.empty((length,), dtype=dt)
                np.copyto(field_arr, data_table_accessor.get_column(data, fname))
            else:
                field_arr = data_table_accessor.get_column(data, fname)

            if self._len is None:
                self._len = len(field_arr)
            elif len(field_arr) != self._len:
                raise ValueError(
                    'All field arrays must have the same length. '
                    f'Field "{fname}" has length {len(field_arr)}, but must be '
                    f'{self._len}!')

            self._data_fields[fname] = field_arr

        if self._len is None:
            # The DataFieldRecordArray is initialized with no fields, i.e. also
            # also no data.
            self._len = 0

        self._field_name_list = list(self._data_fields.keys())
        self._indices = None

    def __contains__(
            self,
            name):
        """Checks if the given field exists in this DataFieldRecordArray
        instance.

        Parameters
        ----------
        name : str
            The name of the field.

        Returns
        -------
        check : bool
            True, if the given field exists in this DataFieldRecordArray
            instance, False otherwise.
        """
        return (name in self._data_fields)

    def __getitem__(
            self,
            name):
        """Implements data field value access.

        Parameters
        ----------
        name : str | numpy ndarray of int or bool
            The name of the data field. If a numpy ndarray is given, it must
            contain the indices for which to retrieve a data selection of the
            entire DataFieldRecordArray. A numpy ndarray of bools can be given
            as well to define a mask.

        Raises
        ------
        KeyError
            If the given data field does not exist.

        Returns
        -------
        data : numpy ndarray | instance of DataFieldRecordArray
            The requested field data or a DataFieldRecordArray holding the
            requested selection of the entire data.
        """
        if isinstance(name, np.ndarray):
            return self.get_selection(name)

        if name not in self._data_fields:
            raise KeyError(
                f'The data field "{name}" is not present in the '
                'DataFieldRecordArray instance.')

        return self._data_fields[name]

    def __setitem__(
            self,
            name,
            arr):
        """Implements data field value assignment. If values are assigned to a
        data field that does not exist yet, it  will be added via the
        ``append_field`` method.

        Parameters
        ----------
        name : str | numpy ndarray of int or bool
            The name of the data field, or a numpy ndarray holding the indices
            or mask of a selection of this DataFieldRecordArray.
        arr : numpy ndarray | instance of DataFieldRecordArray
            The numpy ndarray holding the field values. It must be of the same
            length as this DataFieldRecordArray. If `name` is a numpy ndarray,
            `arr` must be a DataFieldRecordArray.

        Raises
        ------
        ValueError
            If the given data array is not of the same length as this
            DataFieldRecordArray instance.
        """
        if isinstance(name, np.ndarray):
            self.set_selection(name, arr)
            return

        # Check if a new field is supposed to be added.
        if name not in self:
            self.append_field(name, arr)
            return

        # We set a particular already existing data field.
        if len(arr) != self._len:
            raise ValueError(
                f'The length of the to-be-set data ({len(arr)}) must match '
                f'the length ({self._len}) of the DataFieldRecordArray '
                'instance!')

        if not isinstance(arr, np.ndarray):
            raise TypeError(
                'When setting a field directly, the data must be provided as a '
                'numpy ndarray!')

        self._data_fields[name] = arr

    def __len__(self):
        return self._len

    def __sizeof__(self):
        """Calculates the size in bytes of this DataFieldRecordArray instance
        in memory.

        Returns
        -------
        memsize : int
            The memory size in bytes that this DataFieldRecordArray instance
            has.
        """
        memsize = getsizeof([
            self._data_fields,
            self._len,
            self._field_name_list,
            self._indices
        ])
        return memsize

    def __str__(self):
        """Creates a pretty informative string representation of this
        DataFieldRecordArray instance.
        """
        (size, prefix) = get_byte_size_prefix(sys.getsizeof(self))

        max_field_name_len = np.max(
            [len(fname) for fname in self._field_name_list])

        # Generates a pretty string representation of the given field name.
        def _pretty_str_field(name):
            field = self._data_fields[name]
            s = (f'{name.ljust(max_field_name_len)}: '
                 '{'
                 f'dtype: {str(field.dtype)}, '
                 f'vmin: {np.min(field):.3e}, '
                 f'vmax: {np.max(field)}'
                 '}')
            return s

        indent_str = ' '*dsp.INDENTATION_WIDTH
        s = (f'{classname(self)}: {len(self._field_name_list)} fields, '
             f'{len(self)} entries, {np.round(size, 0):.0f} {prefix}bytes ')
        if len(self._field_name_list) > 0:
            s += f'\n{indent_str}fields = '
            s += '{'
            for fname in self._field_name_list:
                s += f'\n{indent_str*2}{_pretty_str_field(fname)}'
            s += f'\n{indent_str}'
            s += '}'

        return s

    @property
    def field_name_list(self):
        """(read-only) The list of the field names of this DataFieldRecordArray.
        """
        return self._field_name_list

    @property
    def indices(self):
        """(read-only) The numpy ndarray holding the indices of this
        DataFieldRecordArray.
        """
        if self._indices is None:
            self._indices = np.arange(self._len)
        return self._indices

    def append(self, arr):
        """Appends the given DataFieldRecordArray to this DataFieldRecordArray
        instance.

        Parameters
        ----------
        arr : instance of DataFieldRecordArray
            The instance of DataFieldRecordArray that should get appended to
            this DataFieldRecordArray. It must contain the same data fields.
            Additional data fields are ignored.
        """
        if not isinstance(arr, DataFieldRecordArray):
            raise TypeError(
                'The arr argument must be an instance of DataFieldRecordArray!')

        for fname in self._field_name_list:
            self._data_fields[fname] = np.append(
                self._data_fields[fname], arr[fname])

        self._len += len(arr)
        self._indices = None

    def append_field(self, name, data):
        """Appends a field and its data to this DataFieldRecordArray instance.

        Parameters
        ----------
        name : str
            The name of the new data field.
        data : numpy ndarray
            The numpy ndarray holding the data. The length of the ndarray must
            match the current length of this DataFieldRecordArray instance.

        Raises
        ------
        KeyError
            If the given data field name already exists in this
            DataFieldRecordArray instance.
        ValueError
            If the length of the data array does not equal to the length of the
            data of this DataFieldRecordArray instance.
        TypeError
            If the arguments are of the wrong type.
        """
        if not isinstance(name, str):
            raise TypeError(
                'The name argument must be an instance of str!')
        if not isinstance(data, np.ndarray):
            raise TypeError(
                'The data argument must be an instance of ndarray!')
        if name in self._data_fields:
            raise KeyError(
                f'The data field "{name}" already exists in this '
                f'{classname(self)} instance!')
        if len(data) != self._len:
            raise ValueError(
                f'The length of the given data is {len(data)}, but must be '
                f'{self._len}!')

        self._data_fields[name] = data
        self._field_name_list.append(name)

    def as_numpy_record_array(self):
        """Creates a numpy record ndarray instance holding the data of this
        DataFieldRecordArray instance.

        Returns
        -------
        arr : instance of numpy record ndarray
            The numpy recarray ndarray holding the data of this
            DataFieldRecordArray instance.
        """
        dt = np.dtype([
            (name, self._data_fields[name].dtype)
            for name in self.field_name_list
        ])

        arr = np.empty((len(self),), dtype=dt)
        for name in self.field_name_list:
            arr[name] = self[name]

        return arr

    def copy(
            self,
            keep_fields=None):
        """Creates a new DataFieldRecordArray that is a copy of this
        DataFieldRecordArray instance.

        Parameters
        ----------
        keep_fields : str | sequence of str | None
            If not None (default), this specifies the data fields that should
            get kept from this DataFieldRecordArray. Otherwise all data fields
            get kept.
        """
        return DataFieldRecordArray(self, keep_fields=keep_fields)

    def remove_field(self, name):
        """Removes the given field from this array.

        Parameters
        ----------
        name : str
            The name of the data field that is to be removed.
        """
        self._data_fields.pop(name)
        self._field_name_list.remove(name)

    def get_field_dtype(self, name):
        """Returns the numpy dtype object of the given data field.
        """
        return self._data_fields[name].dtype

    def set_field_dtype(self, name, dt):
        """Sets the data type of the given field.

        Parameters
        ----------
        name : str
            The name of the data field.
        dt : numpy.dtype
            The dtype instance defining the new data type.
        """
        if name not in self:
            raise KeyError(
                f'The data field "{name}" does not exist in this '
                f'{classname(self)} instance!')
        if not isinstance(dt, np.dtype):
            raise TypeError(
                'The dt argument must be an instance of type numpy.dtype!')

        self._data_fields[name] = self._data_fields[name].astype(dt, copy=False)

    def convert_dtypes(
            self,
            conversions,
            except_fields=None):
        """Converts the data type of the data fields of this
        DataFieldRecordArray. This method can be used to compress the data.

        Parameters
        ----------
        conversions : dict of `old_dtype` -> `new_dtype`
            The dictionary with the old dtype as key and the new dtype as value.
        except_fields : sequence of str | None
            The sequence of field names, which should not get converted.
        """
        if not isinstance(conversions, dict):
            raise TypeError(
                'The conversions argument must be an instance of dict!')

        if except_fields is None:
            except_fields = []
        if not issequenceof(except_fields, str):
            raise TypeError(
                'The except_fields argument must be a sequence of str!')

        _data_fields = self._data_fields
        for fname in self._field_name_list:
            if fname in except_fields:
                continue
            old_dtype = _data_fields[fname].dtype
            if old_dtype in conversions:
                new_dtype = conversions[old_dtype]
                _data_fields[fname] = _data_fields[fname].astype(new_dtype)

    def get_selection(
            self,
            indices):
        """Creates an DataFieldRecordArray that contains a selection of the data
        of this DataFieldRecordArray instance.

        Parameters
        ----------
        indices : (N,)-shaped numpy ndarray of int or bool
            The numpy ndarray holding the indices for which to select the data.

        Returns
        -------
        data_field_array : instance of DataFieldRecordArray
            The DataFieldRecordArray that contains the selection of the
            original DataFieldRecordArray. The selection data is a copy of the
            original data.
        """
        data = dict()
        for fname in self._field_name_list:
            # Get the data selection from the original data. This creates a
            # copy.
            data[fname] = self._data_fields[fname][indices]
        return DataFieldRecordArray(data, copy=False)

    def set_selection(
            self,
            indices,
            arr):
        """Sets a selection of the data of this DataFieldRecordArray instance
        to the data given in arr.

        Parameters
        ----------
        indices : (N,)-shaped numpy ndarray of int or bool
            The numpy ndarray holding the indices or mask for which to set the
            data.
        arr : instance of DataFieldRecordArray
            The instance of DataFieldRecordArray holding the selection data.
            It must have the same fields defined as this DataFieldRecordArray
            instance.
        """
        if not isinstance(arr, DataFieldRecordArray):
            raise TypeError(
                'The arr argument must be an instance of DataFieldRecordArray!')

        for fname in self._field_name_list:
            self._data_fields[fname][indices] = arr[fname]

    def rename_fields(
            self,
            conversions,
            must_exist=False):
        """Renames the given fields of this array.

        Parameters
        ----------
        conversions : dict of `old_name` -> `new_name`
            The dictionary holding the old and new names of the data fields.
        must_exist : bool
            Flag if the given fields must exist. If set to ``True`` and a field
            does not exist, a KeyError is raised.

        Raises
        ------
        KeyError
            If ``must_exist`` is set to ``True`` and a given field does not
            exist.
        """
        for (old_fname, new_fname) in conversions.items():
            if old_fname in self.field_name_list:
                self._data_fields[new_fname] = self._data_fields.pop(old_fname)
            elif must_exist is True:
                raise KeyError(
                    f'The required field "{old_fname}" does not exist!')

        self._field_name_list = list(self._data_fields.keys())

    def tidy_up(
            self,
            keep_fields):
        """Removes all fields that are not specified through the keep_fields
        argument.

        Parameters
        ----------
        keep_fields : str | sequence of str
            The field name(s), that should not be removed.

        Raises
        ------
        TypeError
            If keep_fields is not an instance of str or a sequence of str
            instances.
        """
        if isinstance(keep_fields, str):
            keep_fields = [keep_fields]
        if not issequenceof(keep_fields, str):
            raise TypeError(
                'The keep_fields argument must be a sequence of str!')

        # We need to make a copy of the field_name_list because that list will
        # get changed by the `remove_field` method.
        field_name_list = copy.copy(self._field_name_list)
        for fname in field_name_list:
            if fname not in keep_fields:
                self.remove_field(fname)

    def sort_by_field(
            self,
            name):
        """Sorts the data along the given field name in ascending order.

        Parameters
        ----------
        name : str
            The name of the field along the events should get sorted.

        Returns
        -------
        sorted_idxs : (n_events,)-shaped numpy ndarray
            The numpy ndarray holding the indices of the sorted array.

        Raises
        ------
        KeyError
            If the given data field does not exist.
        """
        if name not in self._data_fields:
            raise KeyError(
                f'The data field "{name}" does not exist in this '
                f'{classname(self)} instance!')

        sorted_idxs = np.argsort(self._data_fields[name])

        for fname in self._field_name_list:
            self._data_fields[fname] = self._data_fields[fname][sorted_idxs]

        return sorted_idxs


register_FileLoader(['.npy'], NPYFileLoader)
register_FileLoader(['.parquet'], ParquetFileLoader)
register_FileLoader(['.pkl'], PKLFileLoader)
register_FileLoader(['.csv'], TextFileLoader)



# -*- coding: utf-8 -*-
# Authors:
#   Dr. Martin Wolf <mail@martin-wolf.org>

import numpy as np

from skyllh.core.debugging import (
    get_logger,
)
from skyllh.core.parameters import (
    ParameterModelMapper,
)
from skyllh.core.pdfratio import (
    SigSetOverBkgPDFRatio,
)
from skyllh.core.py import (
    module_class_method_name,
)


class PDSigSetOverBkgPDFRatio(
        SigSetOverBkgPDFRatio):
    def __init__(
            self,
            sig_pdf_set,
            bkg_pdf,
            cap_ratio=False,
            **kwargs):
        """Creates a PDFRatio instance for the public data.
        It takes a signal PDF set for different discrete gamma values.

        Parameters
        ----------
        sig_pdf_set : instance of PDSignalEnergyPDFSet
            The PDSignalEnergyPDFSet instance holding the set of signal energy
            PDFs.
        bkg_pdf : instance of PDDataBackgroundI3EnergyPDF
            The PDDataBackgroundI3EnergyPDF instance holding the background
            energy PDF.
        cap_ratio : bool
            Switch whether the S/B PDF ratio should get capped where no
            background is available. Default is False.
        """
        self._logger = get_logger(module_class_method_name(self, '__init__'))

        super().__init__(
            sig_pdf_set=sig_pdf_set,
            bkg_pdf=bkg_pdf,
            **kwargs)

        # Construct the instance for the fit parameter interpolation method.
        self._interpolmethod = self.interpolmethod_cls(
            func=self._get_ratio_values,
            param_grid_set=sig_pdf_set.param_grid_set)

        self.cap_ratio = cap_ratio
        if self.cap_ratio:
            self._logger.info('The energy PDF ratio will be capped!')

            # Calculate the ratio value for the phase space where no background
            # is available. We will take the p_sig percentile of the signal
            # like phase space.
            ratio_perc = 99

            # Get the log10 reco energy values where the background pdf has
            # non-zero values.
            n_logE = bkg_pdf.get_binning('log_energy').nbins
            n_sinDec = bkg_pdf.get_binning('sin_dec').nbins
            bd = bkg_pdf._hist_logE_sinDec > 0
            log10_e_bc = bkg_pdf.get_binning('log_energy').bincenters
            self.ratio_fill_value_dict = dict()
            for sig_pdf_key in sig_pdf_set.pdf_keys:
                sigpdf = sig_pdf_set[sig_pdf_key]
                sigvals = sigpdf.get_pd_by_log10_reco_e(log10_e_bc)
                sigvals = np.broadcast_to(sigvals, (n_sinDec, n_logE)).T
                r = sigvals[bd] / bkg_pdf._hist_logE_sinDec[bd]
                # Remove possible inf values.
                r = r[np.invert(np.isinf(r))]
                val = np.percentile(r[r > 1.], ratio_perc)
                self.ratio_fill_value_dict[sig_pdf_key] = val
                self._logger.info(
                    f'The cap value for the energy PDF ratio key {sig_pdf_key} '
                    f'is {val}.')

        # Create cache variables for the last ratio value and gradients in
        # order to avoid the recalculation of the ratio value when the
        # ``get_gradient`` method is called (usually after the ``get_ratio``
        # method was called).
        self._cache_tdm_trial_data_state_id = None
        self._cache_fitparams_hash = None
        self._cache_ratio = None
        self._cache_grads = None

    @property
    def cap_ratio(self):
        """Boolean switch whether to cap the ratio where no background
        information is available (True) or use the smallest possible floating
        point number greater than zero as background pdf value (False).
        """
        return self._cap_ratio

    @cap_ratio.setter
    def cap_ratio(self, b):
        self._cap_ratio = b

    def _is_cached(
            self,
            tdm,
            fitparams_hash):
        """Checks if the ratio and gradients for the given hash of local fit
        parameters are already cached.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The instance of TrialDataManager holding the trial data events.
        fitparams_hach : int
            The hash value of the local fit parameter values.

        Returns
        -------
        check : bool
            ``True`` if the ratio and gradient values are already cached,
            ``False`` otherwise.
        """
        if (self._cache_tdm_trial_data_state_id == tdm.trial_data_state_id) and\
           (self._cache_fitparams_hash == fitparams_hash) and\
           (self._cache_ratio is not None) and\
           (self._cache_grads is not None):
            return True

        return False

    def _get_hash_of_local_sig_fit_param_values(
            self,
            src_params_recarray):
        """Gets the hash of the values of the local signal fit parameters from
        the given ``src_params_recarray``.

        Parameters
        ----------
        src_params_recarray : instance of ndarray
            The (N_sources,)-shaped structured numpy ndarray holding the local
            parameter names and values of the sources.

        Returns
        -------
        hash : int
            The hash of the (N_fitparams, N_sources)-shaped tuple of tuples
            holding the values of the local signal fit parameters.
        """
        values = []
        for param_name in self.sig_param_names:
            if ParameterModelMapper.is_local_param_a_fitparam(
                    param_name, src_params_recarray):
                values.append(tuple(src_params_recarray[param_name]))

        values = tuple(values)

        return hash(values)

    def _get_ratio_values(
            self,
            tdm,
            eventdata,
            gridparams_recarray,
            n_values):
        """Select the signal PDF for the given fit parameter grid point and
        evaluates the S/B ratio for all the trial data events and sources.
        """
        n_sources = len(gridparams_recarray)

        ratio = np.empty((n_values,), dtype=np.double)

        same_pdf_for_all_sources = True
        if len(gridparams_recarray) > 1:
            for pname in gridparams_recarray.dtype.fields.keys():
                if not np.all(np.isclose(np.diff(gridparams_recarray[pname]), 0)):
                    same_pdf_for_all_sources = False
                    break
        if same_pdf_for_all_sources:
            # Special case where the grid parameter values are the same for all
            # sources for all grid parameters
            gridparams = dict(
                zip(gridparams_recarray.dtype.fields.keys(),
                    gridparams_recarray[0])
            )
            sig_pdf_key = self.sig_pdf_set.make_key(gridparams)
            sig_pdf = self.sig_pdf_set.get_pdf(sig_pdf_key)
            (ratio, sig_grads) = sig_pdf.get_pd(
                tdm=tdm,
                params_recarray=None)
        else:
            # General case, we need to loop over the sources.
            for (sidx, interpol_param_values) in enumerate(gridparams_recarray):
                m_src = np.zeros((n_sources), dtype=np.bool_)
                m_src[sidx] = True
                m_values = tdm.get_values_mask_for_source_mask(m_src)

                gridparams = dict(
                    zip(gridparams_recarray.dtype.fields.keys(),
                        interpol_param_values)
                )
                sig_pdf_key = self.sig_pdf_set.make_key(gridparams)
                sig_pdf = self.sig_pdf_set.get_pdf(sig_pdf_key)
                (sig_pd, sig_grads) = sig_pdf.get_pd(
                    tdm=tdm,
                    params_recarray=None)

                ratio[m_values] = sig_pd[m_values]

        (bkg_pd, bkg_grads) = self.bkg_pdf.get_pd(tdm=tdm)
        (bkg_pd,) = tdm.broadcast_selected_events_arrays_to_values_arrays(
            (bkg_pd,))

        m_nonzero_bkg = bkg_pd > 0
        m_zero_bkg = np.invert(m_nonzero_bkg)
        if np.any(m_zero_bkg):
            ev_idxs = np.where(m_zero_bkg)[0]
            self._logger.debug(
                f'For {len(ev_idxs)} events the background probability is '
                f'zero. The event indices of these events are: {ev_idxs}')

        np.divide(
            ratio,
            bkg_pd,
            where=m_nonzero_bkg,
            out=ratio)

        if self._cap_ratio:
            ratio[m_zero_bkg] = self.ratio_fill_value_dict[sig_pdf_key]
        else:
            np.divide(
                ratio,
                np.finfo(np.double).resolution,
                where=m_zero_bkg,
                out=ratio)

        # Check for positive inf values in the ratio and set the ratio to a
        # finite number. Here we choose the maximum value of float32 to keep
        # room for additional computational operations.
        m_inf = np.isposinf(ratio)
        ratio[m_inf] = np.finfo(np.float32).max

        return ratio

    def _calculate_ratio_and_grads(
            self,
            tdm,
            src_params_recarray,
            fitparams_hash):
        """Calculates the ratio and ratio gradient values for all the trial data
        events and sources given the fit parameters using the interpolation
        method for the fit parameter. It caches the results.
        """
        (ratio, grads) = self._interpolmethod(
            tdm=tdm,
            eventdata=None,
            params_recarray=src_params_recarray)

        # Cache the ratio and gradient values.
        self._cache_fitparams_hash = fitparams_hash
        self._cache_ratio = ratio
        self._cache_grads = grads

    def get_ratio(
            self,
            tdm,
            src_params_recarray,
            tl=None):
        """Calculates the PDF ratio values for all events and sources.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The instance of TrialDataManager holding the trial data events for
            which the PDF ratio values should get calculated.
        src_params_recarray : instance of numpy structured ndarray | None
            The (N_sources,)-shaped numpy structured ndarray holding the
            parameter names and values of the sources.
            See the documentation of the
            :meth:`skyllh.core.parameters.ParameterModelMapper.create_src_params_recarray`
            method for more information.
        tl : instance of TimeLord | None
            The optional instance of TimeLord that should be used to measure
            timing information.

        Returns
        -------
        ratios : instance of ndarray
            The (N_values,)-shaped 1d numpy ndarray of float holding the PDF
            ratio value for each trial event and source.
        """
        fitparams_hash = self._get_hash_of_local_sig_fit_param_values(
            src_params_recarray)

        # Check if the ratio value is already cached.
        if self._is_cached(
                tdm=tdm,
                fitparams_hash=fitparams_hash):
            return self._cache_ratio

        self._calculate_ratio_and_grads(
            tdm=tdm,
            src_params_recarray=src_params_recarray,
            fitparams_hash=fitparams_hash)

        return self._cache_ratio

    def get_gradient(
            self,
            tdm,
            src_params_recarray,
            fitparam_id,
            tl=None):
        """Retrieves the PDF ratio gradient for the global fit parameter
        ``fitparam_id`` for each trial data event and source, given the given
        set of parameters ``src_params_recarray`` for each source.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The instance of TrialDataManager holding the trial data events for
            which the PDF ratio gradient values should get calculated.
        src_params_recarray : instance of numpy structured ndarray | None
            The (N_sources,)-shaped numpy structured ndarray holding the
            parameter names and values of the sources.
            See the documentation of the
            :meth:`skyllh.core.parameters.ParameterModelMapper.create_src_params_recarray`
            method for more information.
        fitparam_id : int
            The name of the fit parameter for which the gradient should get
            calculated.
        tl : instance of TimeLord | None
            The optional TimeLord instance that should be used to measure
            timing information.

        Returns
        -------
        grad : instance of ndarray
            The (N_values,)-shaped numpy ndarray holding the gradient values
            for all sources and trial events w.r.t. the given global fit
            parameter.
        """
        fitparams_hash = self._get_hash_of_local_sig_fit_param_values(
            src_params_recarray)

        # Calculate the gradients if they are not calculated yet.
        if not self._is_cached(
            tdm=tdm,
            fitparams_hash=fitparams_hash
        ):
            self._calculate_ratio_and_grads(
                tdm=tdm,
                src_params_recarray=src_params_recarray,
                fitparams_hash=fitparams_hash)

        tdm_n_sources = tdm.n_sources

        grad = np.zeros((tdm.get_n_values(),), dtype=np.float64)

        # Loop through the parameters of the signal PDF set and match them with
        # the global fit parameter.
        for (pidx, pname) in enumerate(
                self._sig_pdf_set.param_grid_set.params_name_list):
            if pname not in src_params_recarray.dtype.fields:
                continue
            p_gpidxs = src_params_recarray[f'{pname}:gpidx']
            src_mask = p_gpidxs == (fitparam_id + 1)
            n_sources = np.count_nonzero(src_mask)
            if n_sources == 0:
                continue
            if n_sources == tdm_n_sources:
                # This parameter applies to all sources, hence to all values,
                # and hence it's the only local parameter contributing to the
                # global parameter fitparam_id.
                return self._cache_grads[pidx]

            # The current parameter does not apply to all sources.
            # Create a values mask that matches a given source mask.
            m_values = tdm.get_values_mask_for_source_mask(src_mask)
            grad[m_values] = self._cache_grads[pidx][m_values]

        return grad



# -*- coding: utf-8 -*-

import itertools
import numpy as np
from copy import deepcopy

from skyllh.core import (
    display,
)
from skyllh.core.model import (
    Model,
    ModelCollection,
)
from skyllh.core.py import (
    NamedObjectCollection,
    bool_cast,
    classname,
    float_cast,
    get_number_of_float_decimals,
    int_cast,
    issequence,
    issequenceof,
)
from skyllh.core.source_model import (
    SourceModel,
)


def make_linear_parameter_grid_1d(
        name,
        low,
        high,
        delta):
    """Utility function to create a ParameterGrid object for a 1-dimensional
    linear parameter grid.

    Parameters
    ----------
    name : str
        The name of the parameter.
    low : float
        The lowest value of the parameter.
    high : float
        The highest value of the parameter.
    delta : float
        The constant distance between the grid values. By definition this
        defines also the precision of the parameter values.

    Returns
    -------
    obj : ParameterGrid
        The ParameterGrid object holding the discrete parameter grid values.
    """
    grid = np.arange(low, high+delta, delta)
    return ParameterGrid(name, grid, delta)


class Parameter(object):
    """This class describes a parameter of a mathematical function, like a PDF,
    or source flux function. A parameter has a name, a value range, and an
    initial value. Furthermore, it has a flag that determines whether this
    parameter has a fixed value or not.
    """
    def __init__(self, name, initial, valmin=None, valmax=None, isfixed=None):
        """Creates a new Parameter instance.

        Parameters
        ----------
        name : str
            The name of the parameter.
        initial : float
            The initial value of the parameter.
        valmin : float | None
            The minimum value of the parameter in case this parameter is
            mutable.
        valmax : float | None
            The maximum value of the parameter in case this parameter is
            mutable.
        isfixed : bool | None
            Flag if the value of this parameter is mutable (False), or not
            (True). If set to `True`, the value of the parameter will always be
            the `initial` value.
            If set to None, the parameter will be mutable if valmin and valmax
            were specified. Otherwise, the parameter is fixed.
            The default is None.
        """
        if isfixed is None:
            if (valmin is not None) and (valmax is not None):
                isfixed = False
            else:
                isfixed = True

        self.name = name
        self.initial = initial
        self.isfixed = isfixed
        self.valmin = valmin
        self.valmax = valmax
        self.value = self._initial

    @property
    def name(self):
        """The name of the parameter.
        """
        return self._name

    @name.setter
    def name(self, name):
        if not isinstance(name, str):
            raise TypeError(
                'The "name" property must be of type str! '
                f'Its current type is {classname(name)}.')
        self._name = name

    @property
    def initial(self):
        """The initial value of the parameter.
        """
        return self._initial

    @initial.setter
    def initial(self, v):
        v = float_cast(
            v,
            'The "initial" property must be castable to type float!')
        self._initial = v

    @property
    def isfixed(self):
        """The flag if the parameter is mutable (False) or not (True).
        """
        return self._isfixed

    @isfixed.setter
    def isfixed(self, b):
        b = bool_cast(
            b,
            'The "isfixed" property must be castable to type bool!')
        self._isfixed = b

    @property
    def valmin(self):
        """The minimum bound value of the parameter.
        """
        return self._valmin

    @valmin.setter
    def valmin(self, v):
        v = float_cast(
            v,
            'The "valmin" property must be castable to type float!',
            allow_None=True)
        self._valmin = v

    @property
    def valmax(self):
        """The maximum bound value of the parameter.
        """
        return self._valmax

    @valmax.setter
    def valmax(self, v):
        v = float_cast(
            v,
            'The "valmax" property must be castable to type float!',
            allow_None=True)
        self._valmax = v

    @property
    def value(self):
        """The current value of the parameter.
        """
        return self._value

    @value.setter
    def value(self, v):
        v = float_cast(
            v,
            'The "value" property must be castable to type float!')
        if self._isfixed:
            if v != self._initial:
                raise ValueError(
                    f'The value ({v}) of the fixed parameter "{self._name}" '
                    'must be equal to the parameter\'s initial value '
                    f'({self._initial})!')
        else:
            if (v < self._valmin) or (v > self._valmax):
                raise ValueError(
                    f'The value ({v}) of parameter "{self._name}" must be '
                    f'within the range [{self._valmin:g}, {self._valmax:g}]!')
        self._value = v

    def __eq__(self, other):
        """Implements the equal comparison operator (==).
        By definition two parameters are equal if there property values are
        equal.

        Parameters
        ----------
        other : Parameter instance
            The instance of Parameter which should be used to compare against
            this Parameter instance.

        Returns
        -------
        cmp : bool
            True, if this Parameter instance and the other Parameter instance
            have the same property values.
        """
        if (self.name != other.name) or\
           (self.value != other.value) or\
           (self.isfixed != other.isfixed):
            return False

        # If both parameters are floating parameters, also their initial, min,
        # and max values must match.
        if not self.isfixed:
            if (self.initial != other.initial) or\
               (self.valmin != other.valmin) or\
               (self.valmax != other.valmax):
                return False

        return True

    def __str__(self):
        """Creates and returns a pretty string representation of this Parameter
        instance.
        """
        indstr = ' ' * display.INDENTATION_WIDTH

        s = f'Parameter: {self._name} = {self._value:g} '

        if self.isfixed:
            s += '[fixed]'
        else:
            s += '[floating] {\n'
            s += indstr + f'initial: {self._initial:g}\n'
            s += indstr + f'range: ({self._valmin:g}, {self._valmax:g})\n'
            s += '}'

        return s

    def as_linear_grid(self, delta):
        """Creates a ParameterGrid instance with a linear grid with constant
        grid value distances delta.

        Parameters
        ----------
        delta : float
            The constant distance between the grid values. By definition this
            defines also the precision of the parameter values.

        Returns
        -------
        grid : ParameterGrid instance
            The ParameterGrid instance holding the grid values.

        Raises
        ------
        ValueError
            If this Parameter instance represents a fixed parameter.
        """

        delta = float_cast(
            delta,
            'The delta argument must be castable to type float!')
        
        if self.isfixed:
            grid = np.array([self.initial])
            return ParameterGrid(self._name, grid, delta)

        grid = make_linear_parameter_grid_1d(
            name=self._name,
            low=self._valmin,
            high=self._valmax,
            delta=delta)

        return grid

    def change_fixed_value(self, value):
        """Changes the value of this fixed parameter to the given value.

        Parameters
        ----------
        value : float
            The parameter's new value.

        Returns
        -------
        value : float
            The parameter's new value.

        Raises
        ------
        ValueError
            If this parameter is not a fixed parameter.
        """
        if not self._isfixed:
            raise ValueError(
                f'The parameter "{self._name}" is not a fixed parameter!')

        self.initial = value
        self.value = value

    def make_fixed(self, initial=None):
        """Fixes this parameter to the given initial value.

        Parameters
        ----------
        initial : float | None
            The new fixed initial value of the Parameter. If set to None, the
            parameter's current value will be used as initial value.

        Returns
        -------
        value : float
            The parameter's new value.
        """
        self._isfixed = True

        # If no new initial value is given, use the current value.
        if initial is None:
            self._initial = self._value
            return self._value

        self.initial = initial
        self._value = self._initial

        # Undefine the valmin and valmax values if the parameter's new value is
        # outside the valmin and valmax range.
        if (self._valmin is not None) and (self._valmax is not None) and\
           ((self._value < self._valmin) or (self._value > self._valmax)):
            self._valmin = None
            self._valmax = None

        return self._value

    def make_floating(self, initial=None, valmin=None, valmax=None):
        """Defines this parameter as floating with the given initial, minimal,
        and maximal value.

        Parameters
        ----------
        initial : float | None
            The initial value of the parameter. If set to `None`, the
            parameter's current value will be used as initial value.
        valmin : float | None
            The minimal value the parameter's value can take.
            If set to `None`, the parameter's current minimal value will be
            used.
        valmax : float | None
            The maximal value the parameter's value can take.
            If set to `None`, the parameter's current maximal value will be
            used.

        Returns
        -------
        value : float
            The parameter's new value.

        Raises
        ------
        ValueError
            If valmin is set to None and this parameter has no valmin defined.
            If valmax is set to None and this parameter has no valmax defined.
        """
        if initial is None:
            initial = self._value
        if valmin is None:
            if self._valmin is None:
                raise ValueError(
                    f'The current minimal value of parameter "{self._name}" '
                    'is not set. So it must be defined through the valmin '
                    'argument!')
            valmin = self._valmin
        if valmax is None:
            if self._valmax is None:
                raise ValueError(
                    f'The current maximal value of parameter "{self._name}" '
                    'is not set. So it must be defined through the valmax '
                    'argument!')
            valmax = self._valmax

        self._isfixed = False
        self.initial = initial
        self.valmin = valmin
        self.valmax = valmax
        self.value = self._initial

        return self._value


class ParameterSet(
        object):
    """This class holds a set of Parameter instances.
    """
    @staticmethod
    def union(*paramsets):
        """Creates a ParameterSet instance that is the union of the given
        ParameterSet instances.

        Parameters
        ----------
        *paramsets : ParameterSet instances
            The sequence of ParameterSet instances.

        Returns
        -------
        paramset : ParameterSet instance
            The newly created ParameterSet instance that holds the union of the
            parameters provided by all the ParameterSet instances.
        """
        if not issequenceof(paramsets, ParameterSet):
            raise TypeError(
                'The arguments of the union static function must be instances '
                'of ParameterSet!')
        if len(paramsets) == 0:
            raise ValueError(
                'At least 1 ParameterSet instance must be provided to the '
                'union static function!')

        paramset = ParameterSet(params=paramsets[0])
        for paramset_i in paramsets[1:]:
            for param in paramset_i._params:
                if not paramset.has_param(param):
                    paramset.add_param(param)

        return paramset

    def __init__(self, params=None):
        """Constructs a new ParameterSet instance.

        Parameters
        ----------
        params : instance of Parameter | sequence of Parameter instances | None
            The initial sequence of Parameter instances of this ParameterSet
            instance.
        """
        # Define the list of parameters.
        # Define the (n_params,)-shaped numpy array of Parameter objects.
        self._params = np.empty((0,), dtype=np.object_)
        # Define the (n_params,)-shaped numpy mask array that masks the fixed
        # parameters in the list of all parameters.
        self._params_fixed_mask = np.empty((0,), dtype=np.bool_)

        # Define two lists for the parameter names. One for the fixed
        # parameters, and one for the floating parameters.
        # This is for optimization purpose only.
        self._fixed_param_name_list = []
        self._floating_param_name_list = []

        # Define dictionaries to map parameter names to storage index of the
        # parameter for fixed and floating parameters.
        self._fixed_param_name_to_idx = dict()
        self._floating_param_name_to_idx = dict()

        # Define a (n_fixed_params,)-shaped ndarray holding the values of the
        # fixed parameters. This is for optimization purpose only.
        self._fixed_param_values = np.empty((0,), dtype=np.float64)

        # Add the initial Parameter instances.
        if params is not None:
            if isinstance(params, Parameter):
                params = [params]
            if not issequenceof(params, Parameter):
                raise TypeError(
                    'The params argument must be None, an instance of '
                    'Parameter, or a sequence of Parameter instances!')
            for param in params:
                self.add_param(param)

    @property
    def params(self):
        """(read-only) The 1D ndarray holding the Parameter instances.
        """
        return self._params

    @property
    def params_name_list(self):
        """(read-only) The list of str holding the names of all the parameters.
        """
        return self._fixed_param_name_list + self._floating_param_name_list

    @property
    def fixed_params(self):
        """(read-only) The 1D ndarray holding the Parameter instances, whose
        values are fixed.
        """
        return self._params[self._params_fixed_mask]

    @property
    def fixed_params_name_list(self):
        """(read-only) The list of the fixed parameter names.
        """
        return self._fixed_param_name_list

    @property
    def fixed_params_mask(self):
        """(read-only) The 1D ndarray holding the mask for the fixed parameters
        of this parameter set.
        """
        return self._params_fixed_mask

    @property
    def fixed_params_idxs(self):
        """The numpy ndarray holding the indices of the fixed parameters.
        """
        idxs = np.argwhere(self._params_fixed_mask).flatten()
        return idxs

    @property
    def floating_params(self):
        """(read-only) The 1D ndarray holding the Parameter instances,
        whose values are floating.
        """
        return self._params[np.invert(self._params_fixed_mask)]

    @property
    def floating_params_name_list(self):
        """(read-only) The list of the floating parameter names.
        """
        return self._floating_param_name_list

    @property
    def floating_params_mask(self):
        """(read-only) The 1D ndarray holding the mask for the floating
        parameters of this parameter set.
        """
        return np.invert(self._params_fixed_mask)

    @property
    def floating_params_idxs(self):
        """The numpy ndarray holding the indices of the floating parameters.
        """
        idxs = np.argwhere(self.floating_params_mask).flatten()
        return idxs

    @property
    def n_params(self):
        """(read-only) The number of parameters this ParameterSet has.
        """
        return len(self._params)

    @property
    def n_fixed_params(self):
        """(read-only) The number of fixed parameters defined in this parameter
        set.
        """
        return len(self._fixed_param_name_list)

    @property
    def n_floating_params(self):
        """(read-only) The number of floating parameters defined in this
        parameter set.
        """
        return len(self._floating_param_name_list)

    @property
    def fixed_param_values(self):
        """(read-only) The (n_fixed_params,)-shaped ndarray holding values of
        the fixed parameters.
        """
        return self._fixed_param_values

    @property
    def floating_param_initials(self):
        """(read-only) The 1D (n_floating_params,)-shaped ndarray holding the
        initial values of all the global floating parameters.
        """
        floating_params = self.floating_params

        if len(floating_params) == 0:
            return np.empty((0,), dtype=np.float64)

        initials = np.array(
            [param.initial for param in floating_params],
            dtype=np.float64)

        return initials

    @property
    def floating_param_bounds(self):
        """(read-only) The 2D (n_floating_params,2)-shaped ndarray holding the
        boundaries for all the floating parameters.
        """
        floating_params = self.floating_params

        if len(floating_params) == 0:
            return np.empty((0, 2), dtype=np.float64)

        bounds = np.array(
            [(param.valmin, param.valmax) for param in floating_params],
            dtype=np.float64)

        return bounds

    def __contains__(self, param_name):
        """Implements the ``param_name in self`` expression. It calls the
        :meth:`has_param` method of this class.

        Parameters
        ----------
        param_name : str
            The name of the parameter.

        Returns
        -------
        check : bool
            Returns ``True`` if the given parameter is part of this ParameterSet
            instance, ``False`` otherwise.
        """
        return self.has_param(param_name)

    def __iter__(self):
        """Returns an iterator over the Parameter instances of this ParameterSet
        instance.
        """
        return iter(self._params)

    def __len__(self):
        """The number of parameters this ParameterSet has.
        """
        return len(self._params)

    def __str__(self):
        """Creates and returns a pretty string representation of this
        ParameterSet instance.
        """
        s = (f'{classname(self)}: {self.n_params} parameters '
             f'({self.n_floating_params} floating, '
             f'{self.n_fixed_params} fixed) ''{')
        for param in self._params:
            s += '\n'
            s += display.add_leading_text_line_padding(
                display.INDENTATION_WIDTH, str(param))
        s += '\n}'
        return s

    def get_fixed_pidx(self, param_name):
        """Returns the parameter index of the given fixed parameter.

        Parameters
        ----------
        param_name : str
            The name of the parameter.

        Returns
        -------
        pidx : int
            The index of the fixed parameter.

        Raises
        ------
        KeyError
            If the given parameter is not part of the set of fixed parameters.
        """
        return self._fixed_param_name_to_idx[param_name]

    def get_floating_pidx(self, param_name):
        """Returns the parameter index of the given floating parameter.

        Parameters
        ----------
        param_name : str
            The name of the parameter.

        Returns
        -------
        pidx : int
            The index of the floating parameter.

        Raises
        ------
        KeyError
            If the given parameter is not part of the set of floating
            parameters.
        """
        return self._floating_param_name_to_idx[param_name]

    def generate_random_floating_param_initials(self, rss):
        """Generates a set of random initials for all floating parameters.
        A new random initial is defined as

            lower_bound + RAND * (upper_bound - lower_bound),

        where RAND is a uniform random variable between 0 and 1.

        Parameters
        ----------
        rss : RandomStateService instance
            The RandomStateService instance that should be used for drawing
            random numbers from.

        Returns
        -------
        ri : (N_floating_params,)-shaped numpy ndarray
            The numpy 1D ndarray holding the generated random initial values.
        """
        vb = self.floating_param_bounds

        # Do random_initial = lower_bound + RAND * (upper_bound - lower_bound).
        ri = (vb[:, 0] +
              rss.random.uniform(size=vb.shape[0])*(vb[:, 1] - vb[:, 0]))

        return ri

    def has_fixed_param(self, param_name):
        """Checks if this ParameterSet instance has a fixed parameter named
        ``param_name``.

        Parameters
        ----------
        param_name : str
            The name of the parameter.

        Returns
        -------
        check : bool
            ``True`` if this ParameterSet instance has a fixed parameter
            of the given name, ``False`` otherwise.
        """
        return (param_name in self._fixed_param_name_list)

    def has_floating_param(self, param_name):
        """Checks if this ParameterSet instance has a floating parameter named
        ``param_name``.

        Parameters
        ----------
        param_name : str
            The name of the parameter.

        Returns
        -------
        check : bool
            ``True`` if this ParameterSet instance has a floating parameter
            of the given name, ``False`` otherwise.
        """
        return (param_name in self._floating_param_name_list)

    def make_params_fixed(self, fix_params):
        """Fixes the given parameters to the given values.

        Parameters
        ----------
        fix_params : dict
            The dictionary defining the parameters that should get fixed to the
            given dictionary entry values.

        Raises
        ------
        ValueError
            If one of the given parameters is already a fixed parameter.
        """
        fix_params_keys = fix_params.keys()
        self._fixed_param_name_list = []
        self._floating_param_name_list = []
        self._fixed_param_name_to_idx = dict()
        self._floating_param_name_to_idx = dict()
        self._fixed_param_values = np.empty((0,), dtype=np.float64)
        for (pidx, param) in enumerate(self._params):
            pname = param.name
            if pname in fix_params_keys:
                # The parameter of name `pname` should get fixed.
                if param.isfixed is True:
                    raise ValueError(
                        f'The parameter "{pname}" is already a fixed '
                        'parameter!')
                initial = fix_params[pname]
                param.make_fixed(initial)
                self._params_fixed_mask[pidx] = True
                self._fixed_param_name_list += [pname]
                self._fixed_param_values = np.concatenate(
                    (self._fixed_param_values, [param.value]))
                self._fixed_param_name_to_idx[pname] = len(
                    self._fixed_param_name_list) - 1
            else:
                if param.isfixed:
                    self._fixed_param_name_list += [pname]
                    self._fixed_param_values = np.concatenate(
                        (self._fixed_param_values, [param.value]))
                    self._fixed_param_name_to_idx[pname] = len(
                        self._fixed_param_name_list) - 1
                else:
                    self._floating_param_name_list += [pname]
                    self._floating_param_name_to_idx[pname] = len(
                        self._floating_param_name_list) - 1

    def make_params_floating(self, float_params):
        """Makes the given parameters floating with the given initial value and
        within the given bounds.

        Parameters
        ----------
        float_params : dict
            The dictionary defining the parameters that should get set to be
            floating. The format of a dictionary's entry can be one of the
            following formats:

                ``None``
                    The parameter's initial, minimal and maximal value should be
                    taken from the parameter's current settings.
                initial : float
                    The parameter's initial value should be set to the given
                    value. The minimal and maximal values of the parameter will
                    be taken from the parameter's current settings.
                (initial, valmin, valmax)
                    The parameter's initial value, minimal and maximal value
                    should be set to the given values. If `initial` is set to
                    `None`, the parameter's current value will be used as
                    initial value.

        Raises
        ------
        ValueError
            If one of the given parameters is already a floating parameter.
        """
        def _parse_float_param_dict_entry(e):
            """Parses the given float_param dictionary entry into initial,
            valmin, and valmax values.
            """
            if e is None:
                return (None, None, None)
            if issequence(e):
                return (e[0], e[1], e[2])
            return (e, None, None)

        float_params_keys = float_params.keys()
        self._fixed_param_name_list = []
        self._floating_param_name_list = []
        self._fixed_param_name_to_idx = dict()
        self._floating_param_name_to_idx = dict()
        self._fixed_param_values = np.empty((0,), dtype=np.float64)
        for (pidx, param) in enumerate(self._params):
            pname = param.name
            if pname in float_params_keys:
                # The parameter of name `pname` should get set floating.
                if param.isfixed is False:
                    raise ValueError(
                        f'The parameter "{pname}" is already a floating '
                        'parameter!')
                (initial, valmin, valmax) = _parse_float_param_dict_entry(
                    float_params[pname])
                param.make_floating(initial, valmin, valmax)
                self._params_fixed_mask[pidx] = False
                self._floating_param_name_list += [pname]
                self._floating_param_name_to_idx[pname] = len(
                    self._floating_param_name_list) - 1
            else:
                if param.isfixed:
                    self._fixed_param_name_list += [pname]
                    self._fixed_param_values = np.concatenate(
                        (self._fixed_param_values, [param.value]))
                    self._fixed_param_name_to_idx[pname] = len(
                        self._fixed_param_name_list) - 1
                else:
                    self._floating_param_name_list += [pname]
                    self._floating_param_name_to_idx[pname] = len(
                        self._floating_param_name_list) - 1

    def update_fixed_param_value_cache(self):
        """Updates the internal cache of the fixed parameter values. This method
        has to be called whenever the values of the fixed Parameter instances
        change.
        """
        for (i, param) in enumerate(self.fixed_params):
            self._fixed_param_values[i] = param.value

    def copy(self):
        """Creates a deep copy of this ParameterSet instance.

        Returns
        -------
        copy : ParameterSet instance
            The copied instance of this ParameterSet instance.
        """
        copy = deepcopy(self)
        return copy

    def add_param(self, param, atfront=False):
        """Adds the given Parameter instance to this set of parameters.

        Parameters
        ----------
        param : instance of Parameter
            The parameter, which should get added.
        atfront : bool
            Flag if the parameter should be added at the front of the parameter
            list. If set to False (default), it will be added at the back.

        Returns
        -------
        self : instance of ParameterSet
            This ParameterSet instance so that multiple add_param calls can just
            be concatenated.

        Raises
        ------
        TypeError
            If param is not an instance of Parameter.
        KeyError
            If given parameter is already present in the set. The check is
            performed based on the parameter name.
        """
        if not isinstance(param, Parameter):
            raise TypeError(
                'The param argument must be an instance of Parameter! '
                f'Its current type is {classname(param)}.')

        if self.has_param(param):
            raise KeyError(
                f'The parameter named "{param.name}" was already added to the '
                'parameter set!')

        param_fixed_mask = True if param.isfixed else False

        if atfront:
            # Add parameter at front of parameter list.
            self._params = np.concatenate(
                ([param], self._params))
            self._params_fixed_mask = np.concatenate(
                ([param_fixed_mask], self._params_fixed_mask))
            if param.isfixed:
                self._fixed_param_name_list = (
                    [param.name] + self._fixed_param_name_list)
                self._fixed_param_values = np.concatenate(
                    ([param.value], self._fixed_param_values))
                # Shift the index of all fixed parameters.
                self._fixed_param_name_to_idx = dict(
                    [(k, v+1)
                     for (k, v) in self._fixed_param_name_to_idx.items()])
                self._fixed_param_name_to_idx[param.name] = 0
            else:
                self._floating_param_name_list = (
                    [param.name] + self._floating_param_name_list)
                # Shift the index of all floating parameters.
                self._floating_param_name_to_idx = dict(
                    [(k, v+1)
                     for (k, v) in self._floating_param_name_to_idx.items()])
                self._floating_param_name_to_idx[param.name] = 0
        else:
            # Add parameter at back of parameter list.
            self._params = np.concatenate(
                (self._params, [param]))
            self._params_fixed_mask = np.concatenate(
                (self._params_fixed_mask, [param_fixed_mask]))
            if param.isfixed:
                self._fixed_param_name_list = (
                    self._fixed_param_name_list + [param.name])
                self._fixed_param_values = np.concatenate(
                    (self._fixed_param_values, [param.value]))
                self._fixed_param_name_to_idx[param.name] = len(
                    self._fixed_param_name_list) - 1
            else:
                self._floating_param_name_list = (
                    self._floating_param_name_list + [param.name])
                self._floating_param_name_to_idx[param.name] = len(
                    self._floating_param_name_list) - 1

        return self

    def has_param(self, param):
        """Checks if the given Parameter is already present in this ParameterSet
        instance. The check is performed based on the parameter name.

        Parameters
        ----------
        param : Parameter instance
            The Parameter instance that should be checked.

        Returns
        -------
        check : bool
            ``True`` if the given parameter is present in this parameter set,
            ``False`` otherwise.
        """
        if (param.name in self._floating_param_name_list) or\
           (param.name in self._fixed_param_name_list):
            return True

        return False

    def get_params_dict(self, floating_param_values):
        """Converts the given floating parameter values into a dictionary with
        the floating parameter names and values and also adds the fixed
        parameter names and their values to this dictionary.

        Parameters
        ----------
        floating_param_values : 1D ndarray
            The ndarray holding the values of the floating parameters in the
            order that the floating parameters are defined.

        Returns
        -------
        params_dict : dict
            The dictionary with the floating and fixed parameter names and
            values.
        """
        params_dict = dict(
            list(zip(self._floating_param_name_list, floating_param_values)) +
            list(zip(self._fixed_param_name_list, self._fixed_param_values))
        )

        return params_dict

    def get_floating_params_dict(self, floating_param_values):
        """Converts the given floating parameter values into a dictionary with
        the floating parameter names and values.

        Parameters
        ----------
        floating_param_values : 1D ndarray
            The ndarray holding the values of the floating parameters in the
            order that the floating parameters are defined.

        Returns
        -------
        params_dict : dict
            The dictionary with the floating and fixed parameter names and
            values.
        """
        params_dict = dict(
            list(zip(self._floating_param_name_list, floating_param_values))
        )

        return params_dict


class ParameterGrid(object):
    """This class provides a data holder for a parameter that has a set of
    discrete values on a grid. Thus, the parameter has a value grid.
    This class represents a one-dimensional grid.
    """
    @staticmethod
    def from_BinningDefinition(binning, delta=None, decimals=None):
        """Creates a ParameterGrid instance from a BinningDefinition instance.

        Parameters
        ----------
        binning : BinningDefinition instance
            The BinningDefinition instance that should be used to create the
            ParameterGrid instance from.
        delta : float | None
            The width between the grid values.
            If set to ``None``, the width is taken from the equal-distant
            ``grid`` values.
        decimals : int | None
            The number of decimals the grid values should get rounded to.
            The maximal number of decimals is 16.
            If set to None, the number of decimals will be the maximum of the
            number of decimals of the first grid value and the number of
            decimals of the delta value.

        Returns
        -------
        param_grid : instance of ParameterGrid
            The created ParameterGrid instance.
        """
        return ParameterGrid(
            name=binning.name,
            grid=binning.binedges,
            delta=delta,
            decimals=decimals)

    @staticmethod
    def from_range(name, start, stop, delta, decimals=None):
        """Creates a ParameterGrid instance from a range definition. The stop
        value will be the last grid point.

        Parameters
        ----------
        name : str
            The name of the parameter grid.
        start : float
            The start value of the range.
        stop : float
            The end value of the range.
        delta : float
            The width between the grid values.
        decimals : int | None
            The number of decimals the grid values should get rounded to.
            The maximal number of decimals is 16.
            If set to None, the number of decimals will be the maximum of the
            number of decimals of the first grid value and the number of
            decimals of the delta value.

        Returns
        -------
        param_grid : instance of ParameterGrid
            The created ParameterGrid instance.
        """
        start = float_cast(
            start,
            'The start argument must be castable to type float!')
        stop = float_cast(
            stop,
            'The stop argument must be castable to type float!')
        delta = float_cast(
            delta,
            'The delta argument must be castable to type float!')
        decimals = int_cast(
            decimals,
            'The decimals argument must be castable to type int!',
            allow_None=True)

        grid = np.arange(start, stop+delta, delta)

        return ParameterGrid(
            name=name,
            grid=grid,
            delta=delta,
            decimals=decimals)

    def __init__(self, name, grid, delta=None, decimals=None):
        """Creates a new parameter grid.

        Parameters
        ----------
        name : str
            The name of the parameter.
        grid : sequence of float
            The sequence of float values defining the discrete grid values of
            the parameter.
        delta : float | None
            The width between the grid values.
            If set to ``None``, the width is taken from the equal-distant
            ``grid`` values.
        decimals : int | None
            The number of decimals the grid values should get rounded to.
            The maximal number of decimals is 16.
            If set to None, the number of decimals will be the maximum of the
            number of decimals of the first grid value and the number of
            decimals of the delta value.
        """
        if delta is None:
            # We need to take the mean of all the "equal" differences in order
            # to smooth out unlucky rounding issues of a particular difference.
            delta = np.mean(np.diff(grid))

        delta = float_cast(
            delta,
            'The delta argument must be castable to type float!')
        self._delta = np.float64(delta)

        # Determine the number of decimals of delta.
        if decimals is None:
            decimals_value = get_number_of_float_decimals(grid[0])
            decimals_delta = get_number_of_float_decimals(delta)
            decimals = int(np.max((decimals_value, decimals_delta)))
        if not isinstance(decimals, int):
            raise TypeError(
                'The decimals argument must be an instance of type int!')
        if decimals > 16:
            raise ValueError(
                'The maximal number of decimals is 16! Maybe you should '
                'consider log-space!?')

        self.name = name
        self._decimals = decimals
        self._delta = np.around(self._delta, self._decimals)
        self.lower_bound = grid[0]

        # Setting the grid, will automatically round the grid values to their
        # next nearest grid value. Hence, we need to set the grid property after
        # setting the delta and offser properties.
        self.grid = grid

    def __str__(self):
        """Pretty string representation.
        """
        return '{:s} = {:s}, decimals = {:d}'.format(
            self._name, str(self._grid), self._decimals)

    @property
    def name(self):
        """The name of the parameter.
        """
        return self._name

    @name.setter
    def name(self, name):
        if not isinstance(name, str):
            raise TypeError(
                'The name property must be of type str!')
        self._name = name

    @property
    def decimals(self):
        """(read-only) The number of significant decimals of the grid values.
        """
        return self._decimals

    @property
    def grid(self):
        """The numpy.ndarray with the grid values of the parameter.
        """
        return self._grid

    @grid.setter
    def grid(self, arr):
        if not issequence(arr):
            raise TypeError(
                'The grid property must be a sequence!')
        if not isinstance(arr, np.ndarray):
            arr = np.array(arr, dtype=np.float64)
        if arr.ndim != 1:
            raise ValueError(
                'The grid property must be a 1D numpy.ndarray!')
        self._grid = self.round_to_nearest_grid_point(arr)

    @property
    def delta(self):
        """(read-only) The width (float) between the grid values.
        """
        return self._delta

    @property
    def lower_bound(self):
        """The lower bound of the parameter grid.
        """
        return self._lower_bound

    @lower_bound.setter
    def lower_bound(self, v):
        v = float_cast(
            v,
            'The lower_bound property must be castable to type float!')
        self._lower_bound = np.around(np.float64(v), self._decimals)

    @property
    def ndim(self):
        """The dimensionality of the parameter grid.
        """
        return self._grid.ndim

    def _calc_floatD_and_intD(self, value):
        """Calculates the number of delta intervals of the given values counted
        from the lower bound of the grid. It returns its float and integer
        representation.
        """
        value = np.atleast_1d(value).astype(np.float64)

        floatD = (value - self._lower_bound)/self._delta
        floatD = np.around(floatD, 9)
        intD = floatD.astype(np.int64)

        return (floatD, intD)

    def add_extra_lower_and_upper_bin(self):
        """Adds an extra lower and upper bin to this parameter grid. This is
        usefull when interpolation or gradient methods require an extra bin on
        each side of the grid.
        """
        newgrid = np.empty((self._grid.size+2,))
        newgrid[1:-1] = self._grid
        newgrid[0] = newgrid[1] - self._delta
        newgrid[-1] = newgrid[-2] + self._delta
        self._lower_bound = newgrid[0]
        del self._grid
        self.grid = newgrid

    def copy(self):
        """Copies this ParameterGrid object and returns the copy.
        """
        copy = deepcopy(self)
        return copy

    def round_to_nearest_grid_point(self, value):
        """Rounds the given value to the nearest grid point.

        Parameters
        ----------
        value : float | ndarray of float
            The value(s) to round.

        Returns
        -------
        grid_point : float | ndarray of float
            The calculated grid point(s).
        """
        scalar_input = np.isscalar(value)

        (floatD, intD) = self._calc_floatD_and_intD(value)
        gp = self._lower_bound + (np.around(floatD % 1, 0) + intD)*self._delta
        gp = np.around(gp, self._decimals)

        if scalar_input:
            return gp.item()

        return gp

    def round_to_lower_grid_point(self, value):
        """Rounds the given value to the nearest grid point that is lower than
        the given value.

        Note: If the given value is a grid point, that grid point will be
              returned!

        Parameters
        ----------
        value : float | ndarray of float
            The value(s) to round.

        Returns
        -------
        grid_point : float | ndarray of float
            The calculated grid point(s).
        """
        scalar_input = np.isscalar(value)

        (floatD, intD) = self._calc_floatD_and_intD(value)
        gp = self._lower_bound + intD*self._delta
        gp = np.around(gp, self._decimals)

        if scalar_input:
            return gp.item()

        return gp

    def round_to_upper_grid_point(self, value):
        """Rounds the given value to the nearest grid point that is larger than
        the given value.

        Note: If the given value is a grid point, the next grid point will be
              returned!

        Parameters
        ----------
        value : float | ndarray of float
            The value(s) to round.

        Returns
        -------
        grid_point : ndarray of float
            The calculated grid point(s).
        """
        scalar_input = np.isscalar(value)

        (floatD, intD) = self._calc_floatD_and_intD(value)
        gp = self._lower_bound + (intD + 1)*self._delta
        gp = np.around(gp, self._decimals)

        if scalar_input:
            return gp.item()

        return gp


class IrregularParameterGrid(object):
    """This class provides a data holder for a parameter that has a set of
    discrete values on a grid. Thus, the parameter has a value grid.
    This class represents a one-dimensional irregular grid.
    """
    @staticmethod
    def from_BinningDefinition(binning):
        """Creates a IrregularParameterGrid instance from a BinningDefinition instance.

        Parameters
        ----------
        binning : BinningDefinition instance
            The BinningDefinition instance that should be used to create the
            ParameterGrid instance from.

        Returns
        -------
        param_grid : instance of ParameterGrid
            The created ParameterGrid instance.
        """
        return IrregularParameterGrid(
            name=binning.name,
            grid=binning.binedges,
        )

    def __init__(self, name, grid):
        """Creates a new parameter grid.

        Parameters
        ----------
        name : str
            The name of the parameter.
        grid : sequence of float
            The sequence of float values defining the discrete grid values of
            the parameter.
        """
        self.name = name
        self.grid = grid

    def __str__(self):
        """Pretty string representation.
        """
        return '{:s} = {:s}'.format(
            self._name, str(self._grid))

    @property
    def name(self):
        """The name of the parameter.
        """
        return self._name

    @name.setter
    def name(self, name):
        if not isinstance(name, str):
            raise TypeError(
                'The name property must be of type str!')
        self._name = name

    @property
    def grid(self):
        """The numpy.ndarray with the grid values of the parameter.
        """
        return self._grid

    @grid.setter
    def grid(self, arr):
        if not issequence(arr):
            raise TypeError(
                'The grid property must be a sequence!')
        if not isinstance(arr, np.ndarray):
            arr = np.array(arr, dtype=np.float64)
        if arr.ndim != 1:
            raise ValueError(
                'The grid property must be a 1D numpy.ndarray!')
        self._grid = arr

    @property
    def ndim(self):
        """The dimensionality of the parameter grid.
        """
        return self._grid.ndim

    def add_extra_lower_and_upper_bin(self):
        """Adds an extra lower and upper bin to this parameter grid. This is
        usefull when interpolation or gradient methods require an extra bin on
        each side of the grid.
        """
        newgrid = np.empty((self._grid.size+2,))
        newgrid[1:-1] = self._grid
        newgrid[0] = newgrid[1] - (newgrid[2] - newgrid[1])
        newgrid[-1] = newgrid[-2] + (newgrid[-2] - newgrid[-3])
        del self._grid
        self.grid = newgrid

    def copy(self):
        """Copies this IrregularParameterGrid object and returns the copy.
        """
        copy = deepcopy(self)
        return copy

    def round_to_nearest_grid_point(self, value):
        """Rounds the given value to the nearest grid point.

        Note: If the given value is precisely in between two nearest grid
        points, the lower nearest grid point will be returned!

        Parameters
        ----------
        value : float | ndarray of float
            The value(s) to round.

        Returns
        -------
        grid_point : float | ndarray of float
            The calculated grid point(s).
        """
        scalar_input = np.isscalar(value)

        grid_middle = (self.grid[1:] + self.grid[:-1])/2
        idx = np.searchsorted(grid_middle, value, side='left')
        gp = self.grid[idx]


        if scalar_input:
            return gp.item()

        return gp

    def round_to_lower_grid_point(self, value):
        """Rounds the given value to the nearest grid point that is lower than
        the given value.

        Note: If the given value is a grid point, that grid point will be
              returned!

        Parameters
        ----------
        value : float | ndarray of float
            The value(s) to round.

        Returns
        -------
        grid_point : float | ndarray of float
            The calculated grid point(s).
        """
        scalar_input = np.isscalar(value)

        idx = np.searchsorted(self.grid, value, side='right') - 1
        gp = self.grid[idx]

        if scalar_input:
            return gp.item()

        return gp

    def round_to_upper_grid_point(self, value):
        """Rounds the given value to the nearest grid point that is larger than
        the given value.

        Note: If the given value is a grid point, the next grid point will be
              returned!

        Parameters
        ----------
        value : float | ndarray of float
            The value(s) to round.

        Returns
        -------
        grid_point : ndarray of float
            The calculated grid point(s).
        """
        scalar_input = np.isscalar(value)

        idx = np.searchsorted(self.grid, value, side='right')
        gp = self.grid[idx]

        if scalar_input:
            return gp.item()

        return gp


class ParameterGridSet(
        NamedObjectCollection):
    """Describes a set of parameter grids.
    """
    def __init__(
            self,
            param_grids=None,
            **kwargs):
        """Constructs a new ParameterGridSet object.

        Parameters
        ----------
        param_grids : sequence of instance of ParameterGrid | instance of ParameterGrid | None
            The ParameterGrid instances this instance of ParameterGridSet should
            get initialized with.
        """
        # Infer `obj_type` from the `param_grids` argument.
        if issequence(param_grids):
            obj_type = type(param_grids[0])
        else:
            obj_type = type(param_grids)

        if obj_type is type(None):
            # Fall back to the default `ParameterGrid` type.
            obj_type = ParameterGrid

        super().__init__(
            objs=param_grids,
            obj_type=obj_type,
            **kwargs)

    @property
    def ndim(self):
        """The dimensionality of this parameter grid set. By definition it's the
        number of parameters of the set.
        """
        return len(self)

    @property
    def params_name_list(self):
        """(read-only) The list of the parameter names.
        """
        return self.name_list

    @property
    def parameter_permutation_dict_list(self):
        """(read-only) The list of parameter dictionaries constructed from all
        permutations of all the parameter values.
        """
        param_grids = [paramgrid.grid for paramgrid in self.objects]

        dict_list = [
            dict([
                (p_i, t_i)
                for (p_i, t_i) in zip(self.name_list, tup)
            ])
            for tup in itertools.product(*param_grids)
        ]

        return dict_list

    def add_extra_lower_and_upper_bin(self):
        """Adds an extra lower and upper bin to all the parameter grids. This is
        usefull when interpolation or gradient methods require an extra bin on
        each side of the grid.
        """
        for paramgrid in self.objects:
            paramgrid.add_extra_lower_and_upper_bin()

    def copy(self):
        """Copies this ParameterGridSet object and returns the copy.
        """
        copy = deepcopy(self)
        return copy


class ParameterModelMapper(
        object):
    """This class provides the parameter to model mapper.
    The parameter to model mapper provides the functionality to map a global
    parameter, usually a fit parameter, to a local parameter of a model, e.g.
    to a source, or a background model parameter.
    """

    @staticmethod
    def is_global_fitparam_a_local_param(
            fitparam_id,
            params_recarray,
            local_param_names):
        """Determines if the given global fit parameter is a local parameter of
        the given list of local parameter names.

        Parameters
        ----------
        fitparam_id : int
            The ID of the global fit parameter.
        params_recarray : instance of numpy record ndarray
            The (N_models,)-shaped numpy record ndarray holding the local
            parameter names and values of the models. See the
            :meth:`skyllh.core.parameters.ParameterModelMapper.create_src_params_recarray`
            method for the format of this record array.
        local_param_names : list of str
            The list of local parameters.

        Returns
        -------
        check : bool
            ``True`` if the global fit parameter translates to a local parameter
            contained in the ``local_param_names`` list, ``False`` otherwise.
        """
        for pname in local_param_names:
            if pname not in params_recarray.dtype.fields:
                continue
            if np.any(params_recarray[f'{pname}:gpidx'] == fitparam_id + 1):
                return True

        return False

    @staticmethod
    def is_local_param_a_fitparam(
            local_param_name,
            params_recarray):
        """Checks if the given local parameter is a (partly) a fit parameter.

        Parameters
        ----------
        local_param_name : str
            The name of the local parameter.
        params_recarray : instance of numpy record ndarray
            The (N_models,)-shaped numpy record ndarray holding the local
            parameter names and values of the models. See the
            :meth:`skyllh.core.parameters.ParameterModelMapper.create_src_params_recarray`
            method for the format of this record array.

        Returns
        -------
        check : bool
            ``True`` if the given local parameter is (partly) a fit parameter.
        """
        if np.any(params_recarray[f'{local_param_name}:gpidx'] > 0):
            return True

        return False

    def __init__(self, models, **kwargs):
        """Constructor of the parameter mapper.

        Parameters
        ----------
        models : sequence of instance of Model.
            The sequence of Model instances the parameter mapper can map global
            parameters to.
        """
        super().__init__(**kwargs)

        models = ModelCollection.cast(
            models,
            'The models property must be castable to an instance of '
            'ModelCollection!')
        self._models = models

        # Create the parameter set for the global parameters.
        self._global_paramset = ParameterSet()

        # Define the attribute holding the boolean mask of the models that are
        # source models.
        self._source_model_mask = np.array(
            [isinstance(model, SourceModel) for model in self._models],
            dtype=bool)

        # Define a (n_models, n_global_params)-shaped numpy ndarray of str
        # objects that will hold the local model parameter names of the global
        # parameters.
        # The local model parameter names are the names used by the internal
        # math objects, like PDFs. Thus, they can be aliases for the global
        # parameter names. Entries set to None, will indicate masked-out
        # global parameters.
        self._model_param_names = np.empty(
            (len(self._models), 0), dtype=np.object_)

    @property
    def models(self):
        """(read-only) The ModelCollection instance defining the models the
        mapper can map global parameters to.
        """
        return self._models

    @property
    def global_paramset(self):
        """(read-only) The ParameterSet instance holding the list of global
        parameters.
        """
        return self._global_paramset

    @property
    def n_models(self):
        """(read-only) The number of models the mapper knows about.
        """
        return len(self._models)

    @property
    def n_global_params(self):
        """(read-only) The number of defined global parameters.
        """
        return self._global_paramset.n_params

    @property
    def n_global_fixed_params(self):
        """(read-only) The number of defined global fixed parameters.
        """
        return self._global_paramset.n_fixed_params

    @property
    def n_global_floating_params(self):
        """(read-only) The number of defined global floating parameters.
        """
        return self._global_paramset.n_floating_params

    @property
    def n_sources(self):
        """(read-only) The number of source models the mapper knows about.
        """
        return np.count_nonzero(self._source_model_mask)

    @property
    def unique_model_param_names(self):
        """(read-only) The unique parameters names of all the models.
        """
        m = self._model_param_names != np.array(None)
        return np.unique(self._model_param_names[m])

    @property
    def unique_source_param_names(self):
        """(read-only) The unique parameter names of the sources.
        """
        src_param_names = self._model_param_names[self._source_model_mask, ...]
        m = src_param_names != np.array(None)
        return np.unique(src_param_names[m])

    def __str__(self):
        """Generates and returns a pretty string representation of this
        parameter model mapper.
        """
        n_global_params = self.n_global_params

        # Determine the number of models that have global parameters assigned.
        # Remember self._model_param_names is a (n_models, n_global_params)-
        # shaped 2D ndarray.
        n_models = self.n_models
        n_sources = self.n_sources

        s = f'{classname(self)}: '
        s += f'{n_global_params} global parameter'
        s += '' if n_global_params == 1 else 's'
        s += ', '
        s += f'{n_models} model'
        s += '' if n_models == 1 else 's'
        s += f' ({n_sources} source'
        s += '' if n_sources == 1 else 's'
        s += ')'

        if n_global_params == 0:
            return s

        s1 = 'Parameters:'
        s += '\n' + display.add_leading_text_line_padding(
            display.INDENTATION_WIDTH, s1)
        for (pidx, p) in enumerate(self._global_paramset.params):
            if p.isfixed:
                pstate = (
                    f'fixed ({p.initial:g})'
                )
            else:
                pstate = (
                    f'floating ({p.valmin:g} <= {p.initial:g} <= {p.valmax:g})'
                )
            ps = f'\n{p.name} [{pstate}]\n'

            ps1 = 'in models:\n'
            for (midx, mpname) in enumerate(self._model_param_names[:, pidx]):
                if mpname is not None:
                    ps1 += '- ' + self._models[midx].name + ': ' + mpname + "\n"

            ps += display.add_leading_text_line_padding(
                display.INDENTATION_WIDTH, ps1)
            s += display.add_leading_text_line_padding(
                2*display.INDENTATION_WIDTH, ps)

        return s

    def get_model_param_name(self, model_idx, gp_idx):
        """Retrieves the local parameter name of a given model and global
        parameter index.

        Parameters
        ----------
        model_idx : int
            The index of the model.
        gp_idx : int
            The index of the global parameter.

        Returns
        -------
        param_name : str | None
            The name of the local model parameter. It is ``None``, if the given
            global parameter is not mapped to the given model.
        """
        param_name = self._model_param_names[model_idx, gp_idx]

        return param_name

    def get_gflp_idx(self, name):
        """Gets the index of the global floating parameter of the given name.

        Parameters
        ----------
        name : str
            The global floating parameter's name.

        Returns
        -------
        idx : int
            The index of the global floating parameter.
        """
        return self._global_paramset.get_floating_pidx(
            param_name=name)

    def get_model_idx_by_name(self, name):
        """Determines the index within this ParameterModelMapper instance of
        the model with the given name.

        Parameters
        ----------
        name : str
            The model's name.

        Returns
        -------
        model_idx : int
            The model's index within this ParameterModelMapper instance.

        Raises
        ------
        KeyError
            If there is no model of the given name.
        """
        for (model_idx, model) in enumerate(self._models):
            if model.name == name:
                return model_idx

        raise KeyError(
            f'The model with name "{name}" does not exist within the '
            'ParameterModelMapper instance!')

    def get_src_model_idxs(self, sources=None):
        """Creates a numpy ndarray holding the indices of the requested source
        models.

        Parameters
        ----------
        sources : instance of SourceModel | sequence of SourceModel | None
            The requested sequence of source models.
            If set to ``None``, all source models will be requested.

        Returns
        -------
        src_model_idxs : numpy ndarray
            The (N_sources,)-shaped 1D ndarray holding the indices of the
            requested source models.
        """
        # Get the model indices of all the source models.
        src_model_idxs = np.arange(self.n_models)[self._source_model_mask]

        if sources is None:
            return src_model_idxs

        # Select only the source models of interest.
        if isinstance(sources, SourceModel):
            sources = [sources]
        if not issequenceof(sources, SourceModel):
            raise TypeError(
                'The sources argument must be None, an instance of '
                'SourceModel, or a sequence of SourceModel! '
                f'Its type is {classname(sources)}')

        src_selection_mask = np.zeros((len(src_model_idxs),), dtype=bool)
        for smidx in src_model_idxs:
            src = self._models[smidx]
            if src in sources:
                src_selection_mask[smidx] = True

        src_model_idxs = src_model_idxs[src_selection_mask]

        return src_model_idxs

    def map_param(self, param, models=None, model_param_names=None):
        """Maps the given instance of Parameter to the given sequence of models
        this parameter model mapper knows about. Aliases for the given parameter
        can be specified for each individual model.

        Parameters
        ----------
        param : instance of Parameter
            The global parameter which should get mapped to one or more models.
        models : sequence of Model instances
            The sequence of Model instances the parameter should get
            mapped to. The instances in the sequence must match Model instances
            specified at construction of this mapper.
        model_param_names : str | sequence of str |  None
            The name of the parameter of the model. Hence, the global
            parameter name can be different to the parameter name of the model.
            If `None`, the name of the global parameter will be used as model
            parameter name for all models.

        Returns
        -------
        self : ParameterModelMapper
            The instance of this ParameterModelMapper, so that several
            `map_param` calls can be concatenated.

        Raises
        ------
        KeyError
            If there is already a model parameter of the same name defined for
            any of the given to-be-applied models.
        """
        if model_param_names is None:
            model_param_names = np.array([param.name]*len(self._models))
        if isinstance(model_param_names, str):
            model_param_names = np.array([model_param_names]*len(self._models))
        if not issequenceof(model_param_names, str):
            raise TypeError(
                'The model_param_names argument must be None, an instance of '
                'str, or a sequence of instances of str!')

        if models is None:
            models = self._models
        models = ModelCollection.cast(
            models,
            'The models argument must be castable to an instance of '
            'ModelCollection!')
        # Make sure that the user did not provide an empty sequence.
        if len(models) == 0:
            raise ValueError(
                'The sequence of models, to which the parameter maps, cannot '
                'be empty!')

        # Get the list of model indices to which the parameter maps.
        mask = np.zeros((self.n_models,), dtype=np.bool_)
        for ((midx, model), applied_model) in itertools.product(
                enumerate(self._models), models):
            if applied_model.id == model.id:
                mask[midx] = True

        # Check that the model parameter name is not already defined for any of
        # the given to-be-mapped models.
        for midx in np.arange(self.n_models)[mask]:
            mpnames = self._model_param_names[midx][
                self._model_param_names[midx] != np.array(None)]
            if model_param_names[midx] in mpnames:
                raise KeyError(
                    f'The model parameter "{model_param_names[midx]}" is '
                    f'already defined for model "{self._models[midx].name}"!')

        self._global_paramset.add_param(param)

        entry = np.where(mask, model_param_names, None)
        self._model_param_names = np.hstack(
            (self._model_param_names, entry[np.newaxis, :].T))

        return self

    def create_model_params_dict(self, gflp_values, model):
        """Creates a dictionary with the fixed and floating parameter names and
        their values for the given model.

        Parameters
        ----------
        gflp_values : 1D ndarray of float
            The ndarray instance holding the current values of the global
            floating parameters.
        model : instance of Model | str | int
            The index of the model as it was defined at construction
            time of this ParameterModelMapper instance.

        Returns
        -------
        model_param_dict : dict
            The dictionary holding the fixed and floating parameter names and
            values of the specified model.
        """
        gflp_values = np.atleast_1d(gflp_values)

        if isinstance(model, str):
            midx = self.get_model_idx_by_name(name=model)
        elif isinstance(model, Model):
            midx = self.get_model_idx_by_name(name=model.name)
        else:
            midx = int_cast(
                model,
                'The model argument must be an instance of Model, str, or '
                'castable to int!')
            if midx < 0 or midx >= len(self._models):
                raise IndexError(
                    f'The model index {midx} is out of range '
                    f'[0,{len(self._models)-1}]!')

        # Get the model parameter mask that masks the global parameters for
        # the requested model.
        m_gp_mask = self._model_param_names[midx] != np.array(None)

        _model_param_names = self._model_param_names
        _global_paramset = self._global_paramset
        gflp_mask = _global_paramset.floating_params_mask
        gfxp_mask = _global_paramset.fixed_params_mask

        # Create the array of local parameter names that belong to the
        # requested model, where the floating parameters are before the fixed
        # parameters.
        model_param_names = np.concatenate((
            _model_param_names[
                midx,
                gflp_mask & m_gp_mask],
            _model_param_names[
                midx,
                gfxp_mask & m_gp_mask]
        ))

        # Create the array of parameter values that belong to the requested
        # model, where floating parameters are before the fixed parameters.
        model_param_values = np.concatenate((
            gflp_values[m_gp_mask[gflp_mask]],
            _global_paramset.fixed_param_values[m_gp_mask[gfxp_mask]]
        ))

        model_param_dict = dict(
            zip(model_param_names, model_param_values))

        return model_param_dict

    def create_src_params_recarray(
            self,
            gflp_values=None,
            sources=None):
        """Creates a numpy record ndarray with a field for each local source
        parameter name and parameter's value. In addition each parameter field
        ``<name>`` has a field named ``<<name>:gpidx>`` which holds the index
        plus one of the corresponding global parameter for each source value.
        For values mapping to fixed parameters, the index is negative. Local
        parameter values that do not apply to a particular source are set to
        NaN. The parameter index in such cases is undefined.
        In addition to the parameter fields, the field ``:model_idx`` holds the
        index of the model for which the local parameter values apply.

        Parameters
        ----------
        gflp_values : numpy ndarray | None
            The (N_global_floating_param,)-shaped 1D ndarray holding the global
            floating parameter values. The order must match the order of
            parameter definition in this ParameterModelMapper instance.
            If set to ``None``, the value ``numpy.nan`` will be used as
            parameter value for floating parameters.
        sources : SourceModel | sequence of SourceModel | ndarray of int32 | None
            The sources which should be considered.
            If a ndarray of type int is provides, it must contain the global
            source indices.
            If set to ``None``, all sources are considered.

        Returns
        -------
        recarray : numpy structured ndarray
            The (N_sources,)-shaped numpy structured ndarray holding the local
            parameter names and their values for each requested source.
            It contains the following fields:

                :model_idx
                    The field holding the index of the model to which the set
                    of local parameters apply.
                <name>
                    The field holding the value for the local parameter <name>.
                    Not all local parameters apply to all sources.
                    Example: "gamma".
                <name>:gpidx
                    The field holding the global parameter index plus one for
                    the local parameter <name>. Example: "gamma:gpidx". Indices
                    for values mapping to fixed parameters are negative.
        """
        if gflp_values is None:
            gflp_values = np.full((self.n_global_floating_params,), np.nan)

        gflp_values = np.atleast_1d(gflp_values)

        # Check input.
        n_global_floating_params = self.n_global_floating_params
        if len(gflp_values) != n_global_floating_params:
            raise ValueError(
                f'The gflp_values argument is of length '
                f'{len(gflp_values)}, but must be of length '
                f'{n_global_floating_params}!')

        if isinstance(sources, np.ndarray) and sources.dtype == np.int32:
            # The sources are already specified in terms of their source
            # indices.
            smidxs = sources
        else:
            # Get the source indices of the requested sources.
            smidxs = self.get_src_model_idxs(sources=sources)

        # Create the output record array with nan as default value.
        dtype = [(':model_idx', np.int32)]
        for name in self.unique_source_param_names:
            dtype += [(name, np.float64), (f'{name}:gpidx', np.int32)]

        recarray = np.zeros(
            (len(smidxs),),
            dtype=dtype)
        for name in self.unique_source_param_names:
            recarray[name] = np.nan

        recarray[':model_idx'] = smidxs

        # Loop over the requested sources.
        _model_param_names = self._model_param_names
        _global_paramset = self._global_paramset
        gflp_mask = _global_paramset.floating_params_mask
        gfxp_mask = _global_paramset.fixed_params_mask
        for (i, smidx) in enumerate(smidxs):
            # Get the mask that selects the global parameters for the requested
            # source.
            src_gp_mask = _model_param_names[smidx] != np.array(None)

            # Create the array of local parameter names that belong to the
            # requested model, where the floating parameters are before the
            # fixed parameters.
            model_param_names = np.concatenate((
                _model_param_names[smidx, gflp_mask & src_gp_mask],
                _model_param_names[smidx, gfxp_mask & src_gp_mask]
            ))

            # Create the array of local parameter values that belong to the
            # requested model, where the floating parameters are before the
            # fixed parameters.
            model_param_values = np.concatenate((
                gflp_values[
                    src_gp_mask[gflp_mask]],
                _global_paramset.fixed_param_values[
                    src_gp_mask[gfxp_mask]]
            ))

            # Create the array of the global parameter indices.
            gpidxs = np.arange(len(_global_paramset))
            model_gp_idxs = np.concatenate((
                gpidxs[gflp_mask & src_gp_mask] + 1,
                -gpidxs[gfxp_mask & src_gp_mask] - 1,
            ))

            # Loop over the local parameters of the source and fill the
            # params record array.
            for (name, value, gpidx) in zip(
                    model_param_names, model_param_values, model_gp_idxs):
                recarray[name][i] = value
                recarray[f'{name}:gpidx'][i] = gpidx

        return recarray

    def create_global_params_dict(self, gflp_values):
        """Converts the given global floating parameter values into a dictionary
        holding the names and values of all floating and fixed parameters.

        Parameters
        ----------
        gflp_values : numpy ndarray
            The (n_global_floating_params,)-shaped 1D numpy ndarray holding the
            values of the global floating parameters.

        Returns
        -------
        params_dict : dict
            The dictionary holding the parameter name and values of all
            floating and fixed parameters.
        """
        params_dict = self._global_paramset.get_params_dict(
            floating_param_values=gflp_values)

        return params_dict

    def create_global_floating_params_dict(self, gflp_values):
        """Converts the given global floating parameter values into a dictionary
        holding the names and values of all floating parameters.

        Parameters
        ----------
        gflp_values : numpy ndarray
            The (n_global_floating_params,)-shaped 1D numpy ndarray holding the
            values of the global floating parameters.

        Returns
        -------
        params_dict : dict
            The dictionary holding the parameter name and values of all
            floating parameters.
        """
        params_dict = self._global_paramset.get_floating_params_dict(
            floating_param_values=gflp_values)

        return params_dict

    def get_local_param_is_global_floating_param_mask(
            self,
            local_param_names,
    ):
        """Checks which local parameter name is mapped to a global floating
        parameter.

        Parameters
        ----------
        local_param_names : sequence of str
            The sequence of the local parameter names to test.

        Returns
        -------
        mask : instance of ndarray
            The (N_local_param_names,)-shaped numpy ndarray holding the mask
            for each local parameter name if it is mapped to a global floating
            parameter.
        """
        mask = np.zeros(len(local_param_names), dtype=np.bool_)

        global_floating_params_idxs = self._global_paramset.floating_params_idxs

        # Get the global parameter indices for each local parameter name.
        for (local_param_idx, local_param_name) in enumerate(local_param_names):
            gpidxs = np.unique(
                np.nonzero(self._model_param_names == local_param_name)[1]
            )
            if np.any(np.isin(gpidxs, global_floating_params_idxs)):
                mask[local_param_idx] = True

        return mask
